! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: ingestsrcnames_aux
!
! Authors:  Masaaki Sakano - SSC/LUX
!           Ian Stewart - SSC/LUX (originally, extractnames_mod.f90)
!
! Description: Module functions to extract catalog source names (IAUNAME), DETID, OBS_ID for a given observation from the source catalogue (such as, 2XMM)  and ingests them, as well as INCATFLG, into the observation-summary source list.
!
! Comments: 
!
! 2006/06 : XMMSRCNAME -> IAUNAME
!           SUM_SRC_NUM -> SRC_NUM
!           OBSID is OK, now
!           Copy DET_ID in the each-obs source file from the Catalogue.
! 2007/05 : The input general catalogue file must have a SRC table 'SRCLIST'
!	that has the columns IAUNAME(XMMSRCNAME), DET_ID(DETID),
!                            SRC_NUM(SUM_SRC_NUM), OBSID(OBS_ID)
!
! TODO:
!  - Rewrite all the functions/subroutines as being based on an Object rather than on multiple variables.
!    It would require to rewrite ingestsrcnames_mod.f90, too.
!  - Change some subroutines into functions, so does perhaps its name (get***).
!
!-
module ingestsrcnames_aux

  use dal
  use errorHandling
!  use datafilter

  use dal_aux			! for readColDataToFixed()
  use ssc_misc
  use array_utils

  implicit none 

  private 

  public allocateSrcPrm
  public initialiseSrcPrm
  public deallocateSrcPrm
  public dumpSrcPrm
  public getObsIdInt32FromHeader
  public getObsIdStr
  public getSrcNumStr
  public getIAUNameStr
  public getDetIdStr
  public getSrcIdStr
  public getInCatFlgStr
  public getRaStr
  public getDecStr
  public getCatBasedPrmsFromSetTabName
  public getCat5Cols
  public getNew3Cols
  public getNew3Or4ColsFromCatColCatData
  public addNew3Cols
  public getColKwdStr	! Public just for the sake of testing...
  public retKwdFromAnyBlock

  interface deallocateSrcPrm
    module procedure deallocateSrcPrmScalar, deallocateSrcPrmAry
  end interface

  interface getCatBasedPrmsFromSetTabName
    module procedure getCatBasedColsFromSetTabName, getCatBasedObjFromSetTabName
  end interface

  interface getObsIdInt32FromHeader
    module procedure getObsIdInt32FromHeaderFromName, getObsIdInt32FromHeaderFromSet
    module procedure getObsIdInt32FromHeaderFromTab
  end interface

  interface getCat5Cols
    module procedure getCat5ColsObjFromSet, getCat5ColsFromSet
    module procedure getCatPrmFromTab
    module procedure getCat5ColsFromTab
  end interface

  interface addNew3Cols
    module procedure addNew3ColsWithCatSetFromSet
    module procedure addNew3ColsWithCatPrmsFromSet
    module procedure addNew3ColsWithCatSetFromTab
    module procedure addNew3ColsWithCatPrmsFromTab
  end interface

  interface getNew3Cols
    module procedure set3To6ElmInSrcFromCatSet
    module procedure getNew3ColsFromCatSet
    module procedure getNew3ColsFromColumnsSId
    module procedure getNew3ColsFromColumnsNoS
    module procedure getNew3ColsFromCatColCatDataSId
    module procedure getNew3ColsFromCatColCatDataNoS
  end interface

  interface getColKwdStr
    module procedure getColKwdStrFromKey, getColKwdStrFromAry
  end interface

  interface retKwdFromAnyBlock
    module procedure retKwdFromAnyBlockInt32FromName, retKwdFromAnyBlockInt32FromSet
    module procedure retKwdFromAnyBlockInt16FromName, retKwdFromAnyBlockInt16FromSet
    module procedure retKwdFromAnyBlockInt8FromName,  retKwdFromAnyBlockInt8FromSet
    module procedure retKwdFromAnyBlockDblFromName,   retKwdFromAnyBlockDblFromSet
    module procedure retKwdFromAnyBlockSglFromName,   retKwdFromAnyBlockSglFromSet
    module procedure retKwdFromAnyBlockCharFromName,  retKwdFromAnyBlockCharFromSet
  end interface

  !.................
  ! Global parameters:

  integer, public, parameter ::&  
     LENINPUTFILENAME = 255 &
   , LENONEFITSLINE = 80 &
   , LENOBSIDSTR = 128 &
   , LENMAXSRCNAME  = 1024 &
   , LENSTRATTR  = 1024 &
   , LENSTRCOLNAME = 255 &
   , LENSTRFILENAME = 1024 &
   , LENGTHOFIAUNAME = 22	! "2XMMp J890123.5-780012"
  character(LENGTHOFIAUNAME), public, parameter ::&
     BLANKSTRING = repeat(' ', LENGTHOFIAUNAME)
  character(7), public, parameter :: DEFAULTSRCTBLNAME = 'SRCLIST'

  !.................
  ! Type definition (public):

  type, public :: SrcPrmFlagAllocT
    logical :: &
         obsId     = .false. &
       , srcNum    = .false. &
       , detid     = .false. &
       , srcid     = .false. &
       , inCatFlag = .false. &
       , catName   = .false. &
       , catNameCol = .false. &
       , raDeg     = .false. &
       , decDeg    = .false.
  end type SrcPrmFlagAllocT

  type, public :: SrcPrmT
    integer(int32), pointer :: obsId(:), srcNum(:), detid(:), srcid(:)
    logical(bool),  pointer :: inCatFlag(:)
    character(LENMAXSRCNAME), pointer :: catName(:)
    type(ColumnT)   :: catNameCol
    real(double),   pointer :: raDeg(:), decDeg(:)
    type(SrcPrmFlagAllocT) :: flagAlloc
    ! NOTE:
    ! 1) Deallocation of this object is done via
    !      call deallocateSrcPrm(srcPrm) 
    !
    ! 2) Once you allocate or define one (eg, obsId), make sure to set
    !      srcPrm%flagAlloc%obsId = .true.
    !    because that information is used whether pointer is set or not
    !    and whether it should be deallocated or not.
    !
    ! 3) Do not write, such as,
    !      srcPrm%srcNum => int32Data(column(catTab, 'SRC_NUM'))
    !      srcPrm%flagAlloc%obsId = .true.
    !    because it will cause a fatal error when trying to deallocate it
    !    via  call deallocateSrcPrm(srcPrm) .
    !    If you don't set  %flagAlloc  in this case, deallocation
    !    will be OK, however that may cause a trouble somewhere else.
    !    Therefore, write instead something like
    !      srcNumData => int32Data(column(catTab, 'SRC_NUM'))
    !      srcPrm%srcNum = srcNumData
    !    that is, actually _copy_ the values to srcPrm .
    !
    ! *) It is debatable whether coping the values is better than
    !    using pointers to columns.  Well, I think it is
    !    more straightforward and intuitive from the coding point of view.
    !    But if you want to use the pointers to columns as opposed to
    !    the current specification, then you may well introduce
    !    another flag parameter, say, flagCol to indicate
    !    whether the pointer of interest should be deallocated or not.
    !    Obviously that will cause a significant amount of changes
    !    in the entire code, for flagCol should be set
    !    appropriately when needed.
  end type SrcPrmT

contains 

  !---------------------------------------------------------------------
  subroutine allocateSrcPrm(srcPrm, numRows, obsId, srcNum, detid, srcid, inCatFlag &
     , catName, raDeg, decDeg, setAllGiven, setInit)
  !---------------------------------------------------------------------
  ! Allocate the specified pointer members of a srcPrm variable.
  ! If setAllGiven is specified and .true. (Default: .false.), then
  !   srcPrm%flagAlloc%GIVEN_FLAG = GIVEN_FLAG
  ! is set; for example, if obsId=.false. is given,
  !   srcPrm%flagAlloc%obsId = obsId	! == .false.
  ! Conversely, if setAllGiven==.false., those pointer members are simply ignored. 
  ! If setInit is specified and .true. (Default: .false.), then
  ! all the flagAlloc members are set even if the FLAG is not given
  ! in the arguments; for example,
  !   call allocateSrcPrm(srcPrm, numRows, obsId=.false.)
  ! sets
  !   srcPrm%flagAlloc%ALL_MEMBERS = .false.
  !
  ! Make sure to  deallocateSrcPrm(srcPrm)  after use.

    implicit none 

    type(SrcPrmT), intent(inout) :: srcPrm
    integer(int32), intent(in) :: numRows
    logical, intent(in), optional ::  obsId, srcNum, detid, srcid, inCatFlag &
       , catName, raDeg, decDeg, setAllGiven, setInit

    logical :: setAllLoc, setInitLoc

    !-------------------------------------------------------------------
    ! Start:

    if (present(setAllGiven)) then
      setAllLoc = setAllGiven
    else
      setAllLoc = .false.	! Default
    end if

    if (present(setInit)) then
      setInitLoc = setInit
    else
      setInitLoc = .false.	! Default
    end if

    if (present(obsId)) then
      if (obsId) then
        allocate( srcPrm%obsId(numRows))
        srcPrm%flagAlloc%obsId     = .true.
      else if (setAllLoc) then
        srcPrm%flagAlloc%obsId     = .false.
      end if
    else if (setInitLoc) then
        srcPrm%flagAlloc%obsId     = .false.
    end if

    if (present(srcNum)) then
      if (srcNum) then
        allocate( srcPrm%srcNum(numRows))
        srcPrm%flagAlloc%srcNum    = .true.
      else if (setAllLoc) then
        srcPrm%flagAlloc%srcNum    = .false.
      end if
    else if (setInitLoc) then
        srcPrm%flagAlloc%srcNum    = .false.
    end if

    if (present(detid)) then
      if (detid) then
        allocate( srcPrm%detid(numRows))
        srcPrm%flagAlloc%detid     = .true.
      else if (setAllLoc) then
        srcPrm%flagAlloc%detid     = .false.
      end if
    else if (setInitLoc) then
        srcPrm%flagAlloc%detid     = .false.
    end if

    if (present(srcid)) then
      if (srcid) then
        allocate( srcPrm%srcid(numRows))
        srcPrm%flagAlloc%srcid     = .true.
      else if (setAllLoc) then
        srcPrm%flagAlloc%srcid     = .false.
      end if
    else if (setInitLoc) then
        srcPrm%flagAlloc%srcid     = .false.
    end if

    if (present(inCatFlag)) then
      if (inCatFlag) then
        allocate( srcPrm%inCatFlag(numRows))
        srcPrm%flagAlloc%inCatFlag = .true.
      else if (setAllLoc) then
        srcPrm%flagAlloc%inCatFlag = .false.
      end if
    else if (setInitLoc) then
        srcPrm%flagAlloc%inCatFlag = .false.
    end if

    if (present(catName)) then
      if (catName) then
        allocate( srcPrm%catName(numRows))
        srcPrm%flagAlloc%catName   = .true.
      else if (setAllLoc) then
        srcPrm%flagAlloc%catName   = .false.
      end if
    else if (setInitLoc) then
        srcPrm%flagAlloc%catName   = .false.
    end if

    if (present(raDeg)) then
      if (raDeg) then
        allocate( srcPrm%raDeg(numRows))
        srcPrm%flagAlloc%raDeg     = .true.
      else if (setAllLoc) then
        srcPrm%flagAlloc%raDeg     = .false.
      end if
    else if (setInitLoc) then
        srcPrm%flagAlloc%raDeg     = .false.
    end if

    if (present(decDeg)) then
      if (decDeg) then
        allocate( srcPrm%decDeg(numRows))
        srcPrm%flagAlloc%decDeg    = .true.
      else if (setAllLoc) then
        srcPrm%flagAlloc%decDeg    = .false.
      end if
    else if (setInitLoc) then
        srcPrm%flagAlloc%decDeg    = .false.
    end if

  end subroutine allocateSrcPrm


  !---------------------------------------------------------------------
  function initialiseSrcPrm(obsId, srcNum, detid, srcid, inCatFlag &
     , catName, catNameCol, raDeg, decDeg) result(srcPrm)
  !---------------------------------------------------------------------
  ! Return an initialised SrcPrmT object.
  ! Note that any component present() here is given .true. for its flag.
  ! Make sure to  deallocateSrcPrm(srcPrm)  after use.

    implicit none 

    type(SrcPrmT) :: srcPrm
    integer(int32), pointer, optional :: obsId(:), srcNum(:), detid(:), srcid(:)
    logical(bool),  pointer, optional :: inCatFlag(:)
    character(*),   pointer, optional :: catName(:)
    type(ColumnT), intent(in), optional :: catNameCol
    real(double),   pointer, optional :: raDeg(:), decDeg(:)

    !-------------------------------------------------------------------
    ! Start:

    if (present(obsId))      then
      srcPrm%obsId => obsId
      srcPrm%flagAlloc%obsId = .true.
    end if
    if (present(srcNum))     then
      srcPrm%srcNum => srcNum
      srcPrm%flagAlloc%srcNum = .true.
    end if
    if (present(detid))      then
      srcPrm%detid => detid
      srcPrm%flagAlloc%detid = .true.
    end if
    if (present(srcid))      then
      srcPrm%srcid => srcid
      srcPrm%flagAlloc%srcid = .true.
    end if
    if (present(inCatFlag))  then
      srcPrm%inCatFlag => inCatFlag
      srcPrm%flagAlloc%inCatFlag = .true.
    end if
    if (present(catName))    then
      srcPrm%catName => catName
      srcPrm%flagAlloc%catName = .true.
    end if
    if (present(catNameCol)) then
      srcPrm%catNameCol =  catNameCol
      srcPrm%flagAlloc%catNameCol = .true.
    end if
    if (present(raDeg))      then
      srcPrm%raDeg => raDeg
      srcPrm%flagAlloc%raDeg = .true.
    end if
    if (present(decDeg))     then
      srcPrm%decDeg => decDeg
      srcPrm%flagAlloc%decDeg = .true.
    end if

  end function initialiseSrcPrm


  !---------------------------------------------------------------------
  subroutine deallocateSrcPrmScalar(srcPrm, iSrc, varName, minVerbosity)
  !---------------------------------------------------------------------
  ! deallocate a SrcPrmT object.
  ! See deallocateSrcPrmAry for detail.

    implicit none 

    type(SrcPrmT), intent(inout) :: srcPrm
    integer(int32), intent(in) :: iSrc
    character(*),   intent(in), optional :: varName
    integer,        intent(in), optional :: minVerbosity

    !-------------------------------------------------------------------
    ! Start:

    call deallocateSrcPrmAry(srcPrm, (/ iSrc /), varName, minVerbosity)

  end subroutine deallocateSrcPrmScalar

  !---------------------------------------------------------------------
  subroutine deallocateSrcPrmAry(srcPrm, iSrcs, varName, minVerbosity)
  !---------------------------------------------------------------------
  ! deallocate a SrcPrmT object.
  ! If (present(iSrcs) .or. present(varName)), and 
  ! if the current verbosity level is equal to or larther than
  ! the minVerbosity (if given), then print messages (maybe for debugging).
  !  iSrcs: (Array of) index numbers of srcPrm to dump its content
  !     (for debugging purposes) before deallocation.
  !  varName: String to print (as a debugging message).
  !
  ! Example: 
  !   call deallocateSrcPrm(catPrm, (/ 1 /), 'catPrm', Noisy)
  ! where this subroutine will dump the the contents of the source 1
  ! (=probably the 2nd source) in catPrm if (verbosity(AppMsg) >= Noisy).

    implicit none 

    type(SrcPrmT), intent(inout) :: srcPrm
    integer(int32), intent(in), optional :: iSrcs(:)
    character(*),   intent(in), optional :: varName
    integer,        intent(in), optional :: minVerbosity

    character(256) :: varNameLoc
    integer(int32) :: i
    logical :: isDebugLoc

    !-------------------------------------------------------------------
    ! Start:

    isDebugLoc = .false.
    if (present(iSrcs) .or. present(varName)) isDebugLoc = .true.
    if (present(minVerbosity)) then
       if (verbosity(AppMsg) >= minVerbosity) then
         ! Do nothing
       else
         isDebugLoc = .false.
       end if
    end if

    if (isDebugLoc .and. present(iSrcs)) then
      do i=1, size(iSrcs)
        call dumpSrcPrm(srcPrm, i)
      end do
    end if

    varNameLoc = ''
    if (present(varName)) then
      varNameLoc = trim(varName)
    end if

    if (isDebugLoc) print *,'----- deallocateSrcPrm: '//trim(varNameLoc)
    if (isDebugLoc .and. srcPrm%flagAlloc%obsId) print *,'Deallocating %obsId'
    if (srcPrm%flagAlloc%obsId)     deallocate(srcPrm%obsId) 
    if (isDebugLoc .and. srcPrm%flagAlloc%srcNum) print *,'Deallocating %srcNum'
    if (srcPrm%flagAlloc%srcNum)    deallocate(srcPrm%srcNum) 
    if (isDebugLoc .and. srcPrm%flagAlloc%detid) print *,'Deallocating %detid'
    if (srcPrm%flagAlloc%detid)     deallocate(srcPrm%detid) 
    if (isDebugLoc .and. srcPrm%flagAlloc%srcid) print *,'Deallocating %srcid'
    if (srcPrm%flagAlloc%srcid)     deallocate(srcPrm%srcid) 
    if (isDebugLoc .and. srcPrm%flagAlloc%inCatFlag) print *,'Deallocating %inCatFlag'
    if (srcPrm%flagAlloc%inCatFlag) deallocate(srcPrm%inCatFlag) 
    if (isDebugLoc .and. srcPrm%flagAlloc%catName) print *,'Deallocating %catName'
    if (srcPrm%flagAlloc%catName)   deallocate(srcPrm%catName) 
    if (isDebugLoc .and. srcPrm%flagAlloc%raDeg) print *,'Deallocating %raDeg'
    if (srcPrm%flagAlloc%raDeg)     deallocate(srcPrm%raDeg) 
    if (isDebugLoc .and. srcPrm%flagAlloc%decDeg) print *,'Deallocating %decDeg'
    if (srcPrm%flagAlloc%decDeg)    deallocate(srcPrm%decDeg) 

    ! %catNameCol had better be untouched, because Table or Set could be released before this call, in which case the following would cause segFault. 
    ! if (isDebugLoc .and. srcPrm%flagAlloc%catNameCol) print *,'Releasing %catNameCol'
    ! if (srcPrm%flagAlloc%catNameCol) call release(srcPrm%catNameCol) 

    if (isDebugLoc) print *,'----- Successfully deallocated.'

  end subroutine deallocateSrcPrmAry


  !---------------------------------------------------------------------
  subroutine dumpSrcPrm(srcPrm, iSrc)
  !---------------------------------------------------------------------
  ! deallocate a SrcPrmT object.

    implicit none 

    type(SrcPrmT), intent(in) :: srcPrm
    integer(int32), intent(in) :: iSrc

    !-------------------------------------------------------------------
    ! Start:

    write(*,'("----- dumpSrcPrm (",i10,")")') iSrc

    if (srcPrm%flagAlloc%obsId) then
      write(*,'("  obsId  =  (", i10, ")")') srcPrm%obsId(iSrc) 
    else
      write(*,'("  obsId  =     F")')
    end if

    if (srcPrm%flagAlloc%srcNum) then
      write(*,'("  srcNum  = (", i10, ")")') srcPrm%srcNum(iSrc)
    else
      write(*,'("  srcNum  =    F")')
    end if

    if (srcPrm%flagAlloc%detid) then
      write(*,'("  detid   = (", i10, ")")') srcPrm%detid(iSrc)
    else
      write(*,'("  detid   =    F")')
    end if

    if (srcPrm%flagAlloc%srcid) then
      write(*,'("  srcid   = (", i10, ")")') srcPrm%srcid(iSrc)
    else
      write(*,'("  srcid   =    F")')
    end if

    if (srcPrm%flagAlloc%inCatFlag) then
      write(*,'("  inCatFlag =  T")')
    else
      write(*,'("  inCatFlag =  F")')
    end if

    if (srcPrm%flagAlloc%catName) then
      write(*,'("  catName   = (", A, ")")') trim(srcPrm%catName(iSrc))
    else
      write(*,'("  catName   =  F")')
    end if

    if (srcPrm%flagAlloc%catNameCol) then
      write(*,'("  catNameCol = T")')
    else
      write(*,'("  catNameCol = F")')
    end if

    if (srcPrm%flagAlloc%raDeg) then
      write(*,'("  raDeg   = (", F15.10, ")")') srcPrm%raDeg(iSrc)
    else
      write(*,'("  raDeg   =    F")')
    end if

    if (srcPrm%flagAlloc%decDeg) then
      write(*,'("  decDeg  = (", F15.10, ")")') srcPrm%decDeg(iSrc)
    else
      write(*,'("  decDeg  =    F")')
    end if

  end subroutine dumpSrcPrm


  !---------------------------------------------------------------------
  subroutine getCatBasedObjFromSetTabName(inCatSetTabName, srcPrm &
     , obsId, getRaDec)
  !---------------------------------------------------------------------
  ! See getCatBasedColsFromSetTabName() for detail.
  ! Make sure to call deallocateSrcPrm(srcPrm) after use.

    implicit none 

    character(*), intent(in) :: inCatSetTabName
    type(SrcPrmT), intent(inout) :: srcPrm
    integer(int32), intent(in), optional :: obsId
    logical, intent(in), optional :: getRaDec	! Default = .true.

    integer(int32) :: obsIdNow, sizeOfSrcPrm
    logical :: getRaDecLoc, noColonFoundCat
    type(DataSetT) :: catSet
    character(LENSTRFILENAME) :: catSetName
    character(LENONEFITSLINE) :: catTabName

    !-------------------------------------------------------------------
    ! Start:

    getRaDecLoc = .false.
    if (present(getRaDec)) getRaDecLoc = getRaDec

    if (present(obsId)) then
      obsIdNow = obsId
    else
      if (.not. srcPrm%flagAlloc%obsId) call error ('Alloc_obsId_getCatBasedObj1','')
      obsIdNow = srcPrm%obsId(lbound(srcPrm%obsId, 1))
    end if

    call splitSetTabName(inCatSetTabName, catSetName, catTabName &
       , noColonFoundCat)	! in ssclib/dal_aux
    if (verbosity(AppMsg) >= Noisy) then
      call message(AppMsg, Noisy, 'Cataglogue name is '//trim(catSetName))
      if (.not. noColonFoundCat) then
        call message(AppMsg, Noisy, 'Cataglogue Table name is '//trim(catTabName))
      else
        call message(AppMsg, Noisy, 'No cataglogue Table name is given.')
      end if
    end if

    if (getRaDecLoc) then
      ! Allocate RA and DEC in srcPrm, if not yet allocated.
      if (.not. srcPrm%flagAlloc%srcNum) call error ('Alloc_srcNum_getCatBasedObj1','')
      sizeOfSrcPrm = size(srcPrm%srcNum)
      if (.not. srcPrm%flagAlloc%raDeg)  then
        allocate(srcPrm%raDeg(sizeOfSrcPrm))
        srcPrm%flagAlloc%raDeg  = .true.
      end if
      if (.not. srcPrm%flagAlloc%decDeg) then
        allocate(srcPrm%decDeg(sizeOfSrcPrm))
        srcPrm%flagAlloc%decDeg = .true.
      end if
    end if

    catSet = dataSet(catSetName, READ)

    if (noColonFoundCat) then
      call set3To6ElmInSrcFromCatSet(catSet, srcPrm &
       , obsId=obsId &
       , useCatCoords=getRaDec)

      ! call getNew3Cols(catSet &
      !    , srcCatName, srcIndex, srcDetid, srcSrcid, srcInCatFlag &
      !    , obsId, catTabName=catTabName &
      !    , srcRaDataD=srcCatRaDeg, srcDecDataD=srcCatDecDeg)
      !    ! srcCatRaDeg and srcCatDecDeg are used ONLY if present().
    else
      call set3To6ElmInSrcFromCatSet(catSet, srcPrm &
       , obsId=obsId, catTabName=catTabName &
       , useCatCoords=getRaDec)

      ! call getNew3Cols(catSet &
      !    , srcCatName, srcIndex, srcDetid, srcSrcid, srcInCatFlag &
      !    , obsId &
      !    , srcRaDataD=srcCatRaDeg, srcDecDataD=srcCatDecDeg)
    end if

    call release(catSet)

    call message(AppMsg, Noisy, 'Closing the catalogue set')

    ! if (getRaDecLoc) then
    !   call getCatBasedColsFromSetTabName(inCatSetTabName, obsIdNow &
    !      , srcPrm%srcNum,  srcPrm%detid,  srcPrm%srcid &
    !      , srcPrm%catName, srcPrm%inCatFlag &
    !      , srcPrm%raDeg,   srcPrm%decDeg)
    ! else
    !   call getCatBasedColsFromSetTabName(inCatSetTabName, obsIdNow &
    !      , srcPrm%srcNum,  srcPrm%detid,  srcPrm%srcid &
    !      , srcPrm%catName, srcPrm%inCatFlag)
    ! end if
    ! 
    ! !! Setup the %flagAlloc
    ! ! srcPrm%srcNum	! Should be preset.
    ! srcPrm%flagAlloc%detid = .true.
    ! srcPrm%flagAlloc%srcid = .true.
    ! srcPrm%flagAlloc%catName   = .true.
    ! srcPrm%flagAlloc%inCatFlag = .true.
    ! if (getRaDecLoc) then
    !   srcPrm%flagAlloc%raDeg  = .true.
    !   srcPrm%flagAlloc%decDeg = .true.
    ! end if

  end subroutine getCatBasedObjFromSetTabName


  !---------------------------------------------------------------------
  subroutine getCatBasedColsFromSetTabName(inCatSetTabName, obsId &
     , srcIndex, srcDetid, srcSrcid &
     , srcCatName, srcInCatFlag &
     , srcCatRaDeg, srcCatDecDeg)
  !---------------------------------------------------------------------
  ! Get Pointer-data of columns (DETID, SRCID, CataloguedName, and
  ! optionally RA and DEC) from the given external catalogue
  ! for the given sources (based on given obsId and SRC_NUM),
  ! and set srcInCatFlag(:) as well.
  ! Note that if you do not want to extract RA and DEC from the catalogue,
  ! do NOT give either srcCatRaDeg or srcCatDecDeg (if you do,
  ! it of course returns the catalogued RA/Dec, which then may
  ! overwrite the existing RA/Dec, depending how you handle the variables).
  !
  ! Make sure to deallocate all the pointers after the call!

    implicit none 

    character(*), intent(in) :: inCatSetTabName
    integer(int32), intent(in) :: obsId
    integer(int32), pointer :: srcIndex(:), srcDetid(:), srcSrcid(:)	! intent(inout)
    character(*),   pointer :: srcCatName(:)	! intent(inout)
    logical(bool),  pointer :: srcInCatFlag(:)	! intent(inout)
    real(double), pointer, optional :: srcCatRaDeg(:), srcCatDecDeg(:)	! intent(inout)
      ! srcCat(Ra|Dec)Deg has the size of sources of interests
      ! and will be filled with the data read from the external catalogue.

    integer(int32), pointer :: obsIdPtr(:)
    type(SrcPrmT) :: srcPrm
    type(DataSetT) :: catSet
    character(LENSTRFILENAME) :: catSetName
    character(LENONEFITSLINE) :: catTabName
    logical :: noColonFoundCat

    !-------------------------------------------------------------------
    ! Start:

    allocate(obsIdPtr(size(srcIndex)))
    obsIdPtr(:) = obsId
    srcPrm = initialiseSrcPrm(obsIdPtr, srcIndex, srcDetid, srcSrcid, srcInCatFlag &
       , srcCatName, raDeg=srcCatRaDeg, decDeg=srcCatDecDeg)	! (ra|dec)Deg may not be present.

    call splitSetTabName(inCatSetTabName, catSetName, catTabName &
       , noColonFoundCat)	! in ssclib/dal_aux
    if (verbosity(AppMsg) >= Noisy) then
      call message(AppMsg, Noisy, 'Cataglogue name is '//trim(catSetName))
      if (.not. noColonFoundCat) then
        call message(AppMsg, Noisy, 'Cataglogue Table name is '//trim(catTabName))
      else
        call message(AppMsg, Noisy, 'No cataglogue Table name is given.')
      end if
    end if

    catSet = dataSet(catSetName, READ)

    if (.not. noColonFoundCat) then
      call getNew3Cols(catSet &
         , srcCatName, srcIndex, srcDetid, srcSrcid, srcInCatFlag &
         , obsId, catTabName=catTabName &
         , srcRaDataD=srcCatRaDeg, srcDecDataD=srcCatDecDeg)
         ! srcCatRaDeg and srcCatDecDeg are used ONLY if present().
    else
      call getNew3Cols(catSet &
         , srcCatName, srcIndex, srcDetid, srcSrcid, srcInCatFlag &
         , obsId &
         , srcRaDataD=srcCatRaDeg, srcDecDataD=srcCatDecDeg)
    end if

    call release(catSet)

    deallocate(obsIdPtr)
    ! NOTE: Do not deallocate(srcPam) as their contents are still being used!

  end subroutine getCatBasedColsFromSetTabName


  !---------------------------------------------------------------------
  subroutine getCat5ColsObjFromSet(catSet, catPrm, catTabName, getRaDec)
  !---------------------------------------------------------------------
  ! See getCat5ColsFromSet() for detail.

    implicit none 

    type(DataSetT), intent(in) :: catSet
    type(SrcPrmT), intent(out) :: catPrm
    character(LENONEFITSLINE), intent(in), optional :: catTabName
    logical, intent(in), optional :: getRaDec	! Default = .true.

    logical :: getRaDecLoc
    type(TableT)   :: catTab

    !-------------------------------------------------------------------
    ! Start:

    getRaDecLoc = .true.
    if (present(getRaDec)) getRaDecLoc = getRaDec

    !! copied from getCat5ColsFromSet() -------------
    if (present(catTabName)) then
      catTab = table(catSet, trim(catTabName))
    else if (blockType(catSet, 0) == TABLE_BLOCK) then
      catTab = table(catSet, 0)
    else if (blockType(catSet, 1) == TABLE_BLOCK) then
      catTab = table(catSet, 1)
    else
      call warning('NoSrcTable' &
         , 'Source table not found in 1st Extension or before.  Try ' &
         //trim(DEFAULTSRCTBLNAME)//'.')
      catTab = table(catSet, trim(DEFAULTSRCTBLNAME))
      !! Or, possibly can read each extension till reaching the 1st Table?
    end if
    !! -------------

    call getCatPrmFromTab(catTab, catPrm &
       , getRaDec=getRaDec)

    ! if (getRaDecLoc) then
    !   call getCat5ColsFromSet(catSet, catPrm%obsId &
    !      , catPrm%srcNum, catPrm%detid, catPrm%srcid, catPrm%catNameCol &
    !      , catTabName=catTabName &
    !      , catRaDataD=catPrm%raDeg, catDecDataD=catPrm%decDeg)
    ! else
    !   call getCat5ColsFromSet(catSet, catPrm%obsId &
    !      , catPrm%srcNum, catPrm%detid, catPrm%srcid, catPrm%catNameCol &
    !      , catTabName=catTabName)
    ! end if
    ! 
    ! !! Setup the %flagAlloc
    ! catPrm%flagAlloc%obsId  = .true.
    ! catPrm%flagAlloc%srcNum = .true.
    ! catPrm%flagAlloc%detid  = .true.
    ! catPrm%flagAlloc%srcid  = .true.
    ! catPrm%flagAlloc%catNameCol  = .true.
    ! if (getRaDecLoc) then
    !   catPrm%flagAlloc%raDeg  = .true.
    !   catPrm%flagAlloc%decDeg = .true.
    ! end if

  end subroutine getCat5ColsObjFromSet


  !---------------------------------------------------------------------
  subroutine getCat5ColsFromSet(catSet, catObsIdDataInt &
     , catSrcNumData, catDetIdData, catSrcIdDataF, catSrcNameCol &
     , catTabName, catRaDataD, catDecDataD)
  !---------------------------------------------------------------------
  ! Get four column pointer arrays.
  ! Make sure to deallocate(catObsIdDataInt, catSrcIdDataF) after use,
  ! as well as (catRaDataD, catDecDataD) if specified, and release(catTab).

    implicit none 

    type(DataSetT), intent(in) :: catSet
    integer(int32), pointer, dimension(:) :: catObsIdDataInt	! intent(out)
    integer(int32), pointer :: catSrcNumData(:), catDetIdData(:), catSrcIdDataF(:)	! intent(out)
    type(ColumnT), intent(out) :: catSrcNameCol
    character(LENONEFITSLINE), intent(in), optional :: catTabName
    real(double), pointer, optional :: catRaDataD(:), catDecDataD(:)

    type(TableT)   :: catTab

    !-------------------------------------------------------------------
    ! Start:

! print *, 'DEBUG: start getCatPrmFromSet()...'
    if (present(catTabName)) then
      catTab = table(catSet, trim(catTabName))
    else if (blockType(catSet, 0) == TABLE_BLOCK) then
      catTab = table(catSet, 0)
    else if (blockType(catSet, 1) == TABLE_BLOCK) then
      catTab = table(catSet, 1)
    else
      call warning('NoSrcTable' &
         , 'Source table not found in 1st Extension or before.  Try ' &
         //trim(DEFAULTSRCTBLNAME)//'.')
      catTab = table(catSet, trim(DEFAULTSRCTBLNAME))
      !! Or, possibly can read each extension till reaching the 1st Table?
    end if

! print *, 'DEBUG: Calling getCat5ColsFromTab()...'
    call getCat5ColsFromTab(catTab, catObsIdDataInt &
     , catSrcNumData, catDetIdData, catSrcIdDataF, catSrcNameCol &
     , catRaDataD=catRaDataD, catDecDataD=catDecDataD)

  end subroutine getCat5ColsFromSet


  !---------------------------------------------------------------------
  subroutine getCatPrmFromTab(catTab, catPrm &
     , getRaDec, getCatName)
  !---------------------------------------------------------------------
  ! Get OBS_ID, SRC_NUM, DETID, SRCID, as well as RA and DEC
  ! and/or catName if requested, plus ColumnT of IAUNAME, all from the catalogue.
  ! Make sure to call, after use
  !    call deallocateSrcPrm(catPrm)
  !    call release(catTab)	! or catSet
  ! Note that the order of these two calls must be as above.
  !
  ! This routine supercedes getCat5ColsFromTab().
  !
  ! NOTE: In this routine, the pointer arrays catTab%...
  !  have the elements with the index starting at 0 instead of 1,
  !  which is not good (in terms of consistency, because the index
  !  of srcPrm everywhere else starts at 1.
  !  However, it is kept as it is for now. JUST BE EXTRA CAREFUL!!!

    implicit none 

    type(TableT), intent(in)   :: catTab
    type(SrcPrmT), intent(out) :: catPrm
    logical, intent(in), optional :: getRaDec, getCatName

    integer(int32), pointer :: catObsIdData(:) &
       , catDetIdDataTmp(:), catSrcNumDataTmp(:) &
       , catSrcIdDataTmp(:)
    real(double), pointer :: catRaDegData(:), catDecDegData(:)
    character(len=LENOBSIDSTR), pointer :: catObsIdDataStr(:)
    character(LEN=LENOBSIDSTR) :: colNameObsIdStr, colNameSrcNumStr &
       , colNameDetIdStr, colNameSrcIdStr, colNameIAUNameStr &
       , colNameRaStr, colNameDecStr
    type(ColumnT) :: obsidCol
    integer(int32) :: nRow
    integer :: dattype, i
    logical :: getRaDecLoc, getCatNameLoc	! .false. in default

    !-------------------------------------------------------------------
    ! Start:

! print *, 'DEBUG: start getCatPrmFromTab()...'

    getRaDecLoc = .false.	! Default
    if (present(getRaDec)) getRaDecLoc = getRaDec
    getCatNameLoc = .false.
    if (present(getCatName)) getCatNameLoc = getCatName

    colNameObsIdStr = getObsIdStr(catTab)
    if (trim(colNameObsIdStr) == '') call error('NoObsIdFoundInCat' &
       , 'No column of OBSID (or alike) is found in the catalogue.')

    nRow = numberOfRows(catTab)-1
    allocate(catPrm%obsId(0:nRow))
    !       , catPrm%srcNum(0:nRow), catPrm%detid(0:nRow))
    !!allocate(catObsIdDataInt(0:nRow), catSrcIdDataF(0:nRow))
    catPrm%flagAlloc%obsId = .true.

    obsidCol = column(catTab, trim(colNameObsIdStr), READ)
    dattype = dataType(obsidCol)
    select case (dattype)
    case (INTEGER32, INTEGER16)
      allocate(catObsIdData(0:nRow))
      catObsIdData => int32Data(obsidCol)
      catPrm%obsId = catObsIdData
      ! catObsIdDataInt = catObsIdData

      deallocate(catObsIdData)

    case (STRING)
      allocate(catObsIdDataStr(0:nRow))

      catObsIdDataStr(:) = ''
      call readColDataToFixed(catTab, trim(colNameObsIdStr),  catObsIdDataStr)
      ! This process takes ages (20min for 2XMMi) for the large catalogue.  Why?

      do i=0,nRow
        read(catObsIdDataStr(i), *) catPrm%obsId(i)
        ! read(catObsIdDataStr(i), *) catObsIdDataInt(i)
      end do
      !!   readColDataToFixed() is defined in dal_aux in ssc_lib,
      !! which is an interface to dal with the use of stringCell() (as below).
      !! Therefore it is very slow...  No proper (and fast) DAL function
      !! exists for this purpose (to read string columns in one go).
      ! do i=0,nRow
      !   cellObsIdStr = stringCell(obsidCol, i)
      !   read(cellObsIdStr, *) catObsIdDataInt(i)
      ! end do

      !catObsIdData => catObsIdDataInt
      deallocate(catObsIdDataStr)

    case default
      call error('WrongDataTypeObsId', 'OBS_ID type is wrong.')
    end select


    ! SRC_NUM
    colNameSrcNumStr  = getSrcNumStr(catTab)
    if (trim(colNameSrcNumStr) == '') then
      call message(AppMsg, Verbose, 'Column SRC_NUM (or alike) is not found in the catalogue.')
      catPrm%flagAlloc%srcNum = .false.
    else
      call message(appMsg, Noisy, 'Column ('//trim(colNameSrcNumStr)//') is found.')
      allocate(catPrm%srcNum(0:nRow))
      catPrm%flagAlloc%srcNum = .true.
      catSrcNumDataTmp => int32Data(column(catTab, colNameSrcNumStr))
      catPrm%srcNum = catSrcNumDataTmp
    end if
    ! colNameSrcNumStr = getSrcNumStr(catTab)
    ! catSrcNumData => int32Data(column(catTab, colNameSrcNumStr))
    ! catPrm%srcNum = catSrcNumData
    ! !catPrm%srcNum => int32Data(column(catTab, colNameSrcNumStr))	! => Causes a fatal error when trying to deallocate it.
    ! !! catSrcNumData => int32Data(column(catTab, colNameSrcNumStr))
    ! catPrm%flagAlloc%srcNum = .true.

    ! DETID
    colNameDetIdStr  = getDetIdStr(catTab)
    if (trim(colNameDetidStr) == '') then
      call message(AppMsg, Verbose, 'Column DETID (or alike) is not found in the catalogue.')
      catPrm%flagAlloc%detid = .false.
    else
      allocate(catPrm%detid(0:nRow))
      catPrm%flagAlloc%detid = .true.
      catDetidDataTmp => int32Data(column(catTab, colNameDetIdStr))
      catPrm%detid = catDetidDataTmp
      call message(appMsg, Noisy, 'Column ('//trim(colNameDetIdStr)//') is found.')
    end if
    ! catDetidData     => int32Data(column(catTab, colNameDetIdStr))
    ! catPrm%detid  = catDetidData
    ! !catPrm%detid     => int32Data(column(catTab, colNameDetIdStr))
    ! !! catDetIdData     => int32Data(column(catTab, colNameDetIdStr))
    ! catPrm%flagAlloc%detid = .true.

    ! SRCID
    colNameSrcIdStr  = getSrcIdStr(catTab, noErr=.true.)
    if (trim(colNameSrcIdStr) == '') then
      call message(AppMsg, Verbose, 'Column SRCID (or alike) is not found in the catalogue.')
      catPrm%flagAlloc%srcid = .false.
      !catPrm%srcid = -1	!! SRCID not exists.
      !! catSrcIdDataF = -1	!! SRCID not exists.
    else
      allocate(catPrm%srcid(0:nRow))
      catPrm%flagAlloc%srcid = .true.
      catSrcIdDataTmp => int32Data(column(catTab, colNameSrcIdStr))
      catPrm%srcid = catSrcIdDataTmp
      ! catSrcIdDataF = catSrcIdDataTmp
      call message(appMsg, Noisy, 'Column ('//trim(colNameSrcIdStr)//') is found.')
    end if

    ! IAUNAME
    colNameIAUNameStr = getIAUNameStr(catTab)
    catPrm%catNameCol = column(catTab, colNameIAUNameStr)
    ! catSrcNameCol = column(catTab, colNameIAUNameStr)
    ! ! arDimCatIAUName = dimensions(catSrcNameCol)
    catPrm%flagAlloc%catNameCol = .true.
    call message(appMsg, Noisy, 'Column ('//trim(colNameIAUNameStr)//') is found.')

    if (getCatNameLoc) then
      ! IAUNAME (as a string)
      allocate(catPrm%catName(0:nRow))
      catPrm%flagAlloc%catName = .true.
      do i=0,nRow
        catPrm%catName(i) = ''	! To avoid funny characters at the end.
        catPrm%catName(i) = stringCell(catPrm%catNameCol, i)
        catPrm%catName(i) = trim(catPrm%catName(i))
      end do

      !!! NOTE !!!
      ! catPrm%catName is NOT filled in default (namely getCatNameLoc==.false.)
      ! because the overhead to do it is significant due to lack
      ! of the handy DAL function to do it.
      ! The above loop does the job only when requested.
      ! In practical cases, most of catPrm%catName do not match srcPrm
      ! (source-list parameters as opposed to the external catalogue, catPrm).
      ! Consequently it is more efficient (in terms of running speed)
      ! to leave catPrm%catName unfilled, but to fill srcPrm%catName
      ! directly only when it is needed in the later stage of
      ! processing.
    end if

    if (getRaDecLoc) then
      ! RA
      allocate(catPrm%raDeg(0:nRow))
      colNameRaStr  = getRaStr(catTab)
      catRaDegData    => real64Data(column(catTab, colNameRaStr))
      catPrm%raDeg    = catRaDegData
      !catPrm%raDeg    => real64Data(column(catTab, colNameRaStr))
      !! catRaDataD    => real64Data(column(catTab, colNameRaStr))
      catPrm%flagAlloc%raDeg = .true.
      call message(appMsg, Noisy, 'Column ('//trim(colNameRaStr)//') is found.')

      ! DEC
      allocate(catPrm%decDeg(0:nRow))
      colNameDecStr = getDecStr(catTab)
      catDecDegData   => real64Data(column(catTab, colNameDecStr))
      catPrm%decDeg   = catDecDegData
      !catPrm%decDeg   => real64Data(column(catTab, colNameDecStr))
      !! catDecDataD   => real64Data(column(catTab, colNameDecStr))
      catPrm%flagAlloc%decDeg = .true.
      call message(appMsg, Noisy, 'Column ('//trim(colNameDecStr)//') is found.')
    end if

  end subroutine getCatPrmFromTab


  !---------------------------------------------------------------------
  subroutine getCat5ColsFromTab(catTab, catObsIdDataInt &
     , catSrcNumData, catDetIdData, catSrcIdDataF, catSrcNameCol &
     , catRaDataD, catDecDataD)
  !---------------------------------------------------------------------
  ! This routine is now obsolete and is superceded by getCatPrmFromTab(),
  ! although it is still used from the routines related to ingestsrcnames .
  ! The main difference is that this routine returns the column pointers,
  ! but not actual arrays as in getCatPrmFromTab().
  !
  ! Get four column pointer arrays and one ColumnT for source-name.
  ! If SRCID does not exist, the values in the returned array are negative (-1).
  ! Make sure to deallocate(catObsIdDataInt, catSrcIdDataF) after use,
  ! as well as (catRaDataD, catDecDataD) if specified, and release(catTab).

    implicit none 

    type(TableT), intent(in)   :: catTab
    integer(int32), pointer :: catObsIdDataInt(:)	! intent(out)
    integer(int32), pointer :: catSrcNumData(:), catDetIdData(:), catSrcIdDataF(:)	! intent(out)
    type(ColumnT), intent(out) :: catSrcNameCol
    real(double), pointer, optional :: catRaDataD(:), catDecDataD(:)

    integer(int32), pointer :: catObsIdData(:), catSrcIdDataTmp(:)
    !character(len=10), pointer :: catObsIdDataStr(:)
    character(len=LENOBSIDSTR), pointer :: catObsIdDataStr(:)
    character(LEN=LENOBSIDSTR) :: colNameObsIdStr, colNameSrcNumStr &
       , colNameDetIdStr, colNameSrcIdStr, colNameIAUNameStr &
       , colNameRaStr, colNameDecStr
    type(ColumnT) :: obsidCol
    integer :: dattype, nrow, i

    !-------------------------------------------------------------------
    ! Start:

    colNameObsIdStr = getObsIdStr(catTab)

    nrow = numberOfRows(catTab)-1
    allocate(catObsIdDataInt(0:nrow), catSrcIdDataF(0:nrow))

    obsidCol = column(catTab, trim(colNameObsIdStr), READ)
    dattype = dataType(obsidCol)
    select case (dattype)
    case (INTEGER32, INTEGER16)
      allocate(catObsIdData(0:nrow))
      catObsIdData => int32Data(obsidCol)
      catObsIdDataInt = catObsIdData

    case (STRING)
      allocate(catObsIdDataStr(0:nrow))

      call readColDataToFixed(catTab, trim(colNameObsIdStr),  catObsIdDataStr)
      ! This process takes ages (20min for 2XMMi) for the large catalogue.  Why?

      do i=0,nrow
        read(catObsIdDataStr(i), *) catObsIdDataInt(i)
      end do
      !!   readColDataToFixed() is defined in dal_aux in ssc_lib,
      !! which is an interface to dal with the use of stringCell() (as below).
      !! Therefore it is very slow...  No proper (and fast) DAL function
      !! exists for this purpose (to read string columns in one go).
      ! do i=0,nrow
      !   cellObsIdStr = stringCell(obsidCol, i)
      !   read(cellObsIdStr, *) catObsIdDataInt(i)
      ! end do

      !catObsIdData => catObsIdDataInt
      deallocate(catObsIdDataStr)

    case default
      call error('WrongDataTypeObsId', 'OBS_ID type is wrong.')
    end select

    colNameSrcNumStr = getSrcNumStr(catTab)
    call message(appMsg, Noisy, 'Column ('//trim(colNameSrcNumStr)//') is found.')
    if (dataType(column(catTab, colNameSrcNumStr)) == INTEGER32) then
      catSrcNumData => int32Data(column(catTab, colNameSrcNumStr))
    else
      call error('wrongDataType', 'Column ('//trim(colNameSrcNumStr)//') is not INTEGER32.')
    end if

    colNameDetIdStr  = getDetIdStr(catTab)
    call message(appMsg, Noisy, 'Column ('//trim(colNameDetIdStr)//') is found.')
    catDetIdData     => int32Data(column(catTab, colNameDetIdStr))

    colNameSrcIdStr  = getSrcIdStr(catTab, noErr=.true.)
    if (trim(colNameSrcIdStr) == '') then
      call message(AppMsg, Noisy, 'Column SRCID (or alike) is not found.')
      catSrcIdDataF = -1	!! SRCID not exists.
    else
      call message(appMsg, Noisy, 'Column ('//trim(colNameSrcIdStr)//') is found.')
      catSrcIdDataTmp => int32Data(column(catTab, colNameSrcIdStr))
      catSrcIdDataF = catSrcIdDataTmp
    end if

    colNameIAUNameStr = getIAUNameStr(catTab)
    call message(appMsg, Noisy, 'Column ('//trim(colNameIAUNameStr)//') is found.')
    catSrcNameCol = column(catTab, colNameIAUNameStr)
    ! arDimCatIAUName = dimensions(catSrcNameCol)

    if (present(catRaDataD)) then
      colNameRaStr  = getRaStr(catTab)
      call message(appMsg, Noisy, 'Column ('//trim(colNameRaStr)//') is found.')
      catRaDataD    => real64Data(column(catTab, colNameRaStr))
    end if
    if (present(catDecDataD)) then
      colNameDecStr = getDecStr(catTab)
      call message(appMsg, Noisy, 'Column ('//trim(colNameDecStr)//') is found.')
      catDecDataD   => real64Data(column(catTab, colNameDecStr))
    end if

  end subroutine getCat5ColsFromTab


  !---------------------------------------------------------------------
  subroutine addNew3ColsWithCatSetFromSet(sumSet, catSet, obsId &
     , sumTabName, catTabName, catRaDataD, catDecDataD)
  !---------------------------------------------------------------------
  ! Add three new columns in the (summary) source list in the given obs.
  ! Make sure to do release(sumSet) after use,
  ! and deallocate(catRaDataD, catDecDataD) if specified.

    implicit none 

    type(DataSetT), intent(in) :: sumSet, catSet
    integer(int32), intent(in), optional :: obsId
    character(LENONEFITSLINE), intent(in), optional :: sumTabName, catTabName
    real(double), pointer, optional :: catRaDataD(:), catDecDataD(:)

    integer(int32), pointer, dimension(:) :: catObsIdDataInt
    integer(int32), pointer :: catSrcNumData(:), catDetIdData(:), catSrcIdDataF(:)
    type(ColumnT) :: catSrcNameCol

    !-------------------------------------------------------------------
    ! Start:

! print *, 'DEBUG: start addNew3ColsWithCatSetFromSet()...'
    call getCat5ColsFromSet(catSet, catObsIdDataInt &
         , catSrcNumData, catDetIdData, catSrcIdDataF &
         , catSrcNameCol, catTabName &
         , catRaDataD, catDecDataD)	! catObsIdDataInt and catSrcIdDataF are allocated.

    call addNew3ColsWithCatPrmsFromSet(sumSet &
         , catObsIdDataInt &
         , catSrcNumData, catDetIdData, catSrcIdDataF, catSrcNameCol &
         , obsId, sumTabName)

    deallocate(catObsIdDataInt, catSrcIdDataF)

  end subroutine addNew3ColsWithCatSetFromSet


  !---------------------------------------------------------------------
  subroutine addNew3ColsWithCatPrmsFromSet(sumSet &
       , catObsIdDataInt &
       , catSrcNumData, catDetIdData, catSrcIdData, catSrcNameCol &
       , obsId, sumTabName)
    !---------------------------------------------------------------------
    ! Add three new columns in the (summary) source list in the given obs.
    ! Make sure to do release(sumSet) after use.

    implicit none 

    type(DataSetT), intent(in) :: sumSet
    integer(int32), pointer, dimension(:) :: catObsIdDataInt	! intent(in)
    integer(int32), pointer :: catSrcNumData(:), catDetIdData(:), catSrcIdData(:)	! intent(in)
    type(ColumnT), intent(in) :: catSrcNameCol
    integer(int32), intent(in), optional :: obsId
    character(LENONEFITSLINE), intent(in), optional :: sumTabName

    type(TableT)  :: sumTab
    integer(int32) :: obsIdNow

    !-------------------------------------------------------------------
    ! Start:

    if (present(obsId)) then
       obsIdNow = obsId
    else
       obsIdNow = getObsIdInt32FromHeaderFromSet(sumSet)
    end if

    if (present(sumTabName)) then
      sumTab = table(sumSet, trim(sumTabName))
    else
      sumTab = table(sumSet, trim(DEFAULTSRCTBLNAME))		! Default
    end if

    call addNew3ColsWithCatPrmsFromTab(sumTab &
         , catObsIdDataInt &
         , catSrcNumData, catDetIdData, catSrcIdData, catSrcNameCol &
         , obsIdNow)

  end subroutine addNew3ColsWithCatPrmsFromSet


  !---------------------------------------------------------------------
  subroutine addNew3ColsWithCatSetFromTab(sumTab, catSet, obsId &
     , catTabName, catRaDataD, catDecDataD)
  !---------------------------------------------------------------------
  ! Add three new columns in the (summary) source list in the given obs.
  ! Make sure to do deallocate(catRaDataD, catDecDataD) if specified.

    implicit none 

    type(TableT), intent(in)   :: sumTab
    type(DataSetT), intent(in) :: catSet
    integer(int32), intent(in), optional :: obsId
    character(LENONEFITSLINE), intent(in), optional :: catTabName
    real(double), pointer, optional :: catRaDataD(:), catDecDataD(:)

    integer(int32), pointer, dimension(:) :: catObsIdDataInt
    integer(int32), pointer :: catSrcNumData(:), catDetIdData(:), catSrcIdDataF(:)
    type(ColumnT) :: catSrcNameCol

    !-------------------------------------------------------------------
    ! Start:

    call getCat5ColsFromSet(catSet, catObsIdDataInt &
       , catSrcNumData, catDetIdData, catSrcIdDataF, catSrcNameCol &
       , catTabName, catRaDataD, catDecDataD)	! catObsIdDataInt and catSrcIdDataF are allocated.

    call addNew3ColsWithCatPrmsFromTab(sumTab &
       , catObsIdDataInt &
       , catSrcNumData, catDetIdData, catSrcIdDataF, catSrcNameCol &
       , obsId)

    deallocate(catObsIdDataInt, catSrcIdDataF)

  end subroutine addNew3ColsWithCatSetFromTab


  !---------------------------------------------------------------------
  subroutine addNew3ColsWithCatPrmsFromTab(sumTab &
     , catObsIdDataInt &
     , catSrcNumData, catDetIdData, catSrcIdData, catSrcNameCol &
     , obsId)
  !---------------------------------------------------------------------
  ! Add three new columns in the (summary) source list in the given obs.

    implicit none 

    type(TableT), intent(in)   :: sumTab
    integer(int32), pointer, dimension(:) :: catObsIdDataInt	! intent(in)
    integer(int32), pointer :: catSrcNumData(:), catDetIdData(:), catSrcIdData(:)	! intent(in)
    type(ColumnT), intent(in) :: catSrcNameCol
    integer(int32), intent(in), optional :: obsId

    integer(int32) :: obsIdNow
    type(ColumnT) :: sumCatSrcNameCol
    integer(int32), pointer :: sumSrcNumData(:), sumDetIdData(:), sumSrcIdData(:)
    logical(bool), pointer :: sumInCatData(:)
    integer :: arDimCatIAUName(0:0), arDimSumIAUName(0:0)
    logical :: useSrcId = .true.

    !-------------------------------------------------------------------
    ! Start:

    if (present(obsId)) then
       obsIdNow = obsId
    else
       obsIdNow = getObsIdInt32FromHeaderFromTab(sumTab)
    end if

    sumSrcNumData => int32Data(column(sumTab, 'SRC_NUM'))

    if (hasColumn(sumTab, 'IAUNAME')) then	! Force 'IAUNAME'
       sumCatSrcNameCol = column(   sumTab, 'IAUNAME')
       arDimSumIAUName = dimensions(sumCatSrcNameCol)
       arDimCatIAUName = dimensions(catSrcNameCol)
       if (arDimSumIAUName(0) < arDimCatIAUName(0)) then
          call warning('smallerDimensionForSumIauname',&
               'IAUNAME in output Summary may be trimmed (due to shortage of the cell size).')
       end if

    else
       arDimSumIAUName(0) = LENGTHOFIAUNAME
       sumCatSrcNameCol = addColumn(sumTab, 'IAUNAME', STRING, &
            comment='Source name', dimensions=arDimSumIAUName)
    end if

    if (hasColumn(sumTab, 'DETID')) then	! Force 'DETID'
       sumDetIdData => int32Data(column(   sumTab, 'DETID'))
    else
       sumDetIdData => int32Data(addColumn(sumTab, 'DETID', INTEGER32, &
            comment='Detection-ID'))
       sumDetIdData = -99			! Initial value of -99
    end if

    useSrcId = .true.
    if (hasColumn(sumTab, 'SRCID')) then	! Force 'SRCID'
       sumSrcIdData => int32Data(column(   sumTab, 'SRCID'))
    else if (catSrcIdData(1) .ge. 0) then
       sumSrcIdData => int32Data(addColumn(sumTab, 'SRCID', INTEGER32, &
            comment='Unique-Source-ID'))
       sumSrcIdData = -99			! Initial value of -99
    else
       useSrcId = .false.
    end if

    if (hasColumn(sumTab, 'INCATFLG')) then
       sumInCatData => boolData(column(   sumTab, 'INCATFLG'))
    else
       sumInCatData => boolData(addColumn(sumTab, 'INCATFLG', BOOLEAN, &
            comment='Whether the source is listed in the catalogue'))
    end if

    if (useSrcId) then
      call getNew3ColsFromColumnsSId(catObsIdDataInt &
         , catSrcNumData, catDetIdData, catSrcIdData, catSrcNameCol &
         , sumCatSrcNameCol, sumSrcNumData, sumDetIdData, sumSrcIdData, sumInCatData &
         , obsIdNow)
    else
      call getNew3ColsFromColumnsNoS(catObsIdDataInt &
         , catSrcNumData, catDetIdData, catSrcIdData, catSrcNameCol &
         , sumCatSrcNameCol, sumSrcNumData, sumDetIdData, sumInCatData &
         , obsIdNow)
    end if

  end subroutine addNew3ColsWithCatPrmsFromTab


  !---------------------------------------------------------------------
  subroutine set3To6ElmInSrcFromCatSet(catSet, srcPrm &
     , obsId, catTabName, useCatCoords)
  !---------------------------------------------------------------------
  ! Set columns of DETID, IAUNAME, INCATFLG, and optionally
  ! SRCID, RA, DEC in the (summary) source list (SrcPrmT object),
  ! using the external catalogue (catSet).
  ! Almost the same as getNew3ColsFromCatSet() but
  ! with different arguments (and supports for RA and DEC).
  ! Make sure to
  !    call deallocateSrcPrm(srcPrm)
  !    call release(catSet)
  ! after use.

    implicit none 

    type(DataSetT), intent(in) :: catSet
    type(SrcPrmT),  intent(inout) :: srcPrm
    integer(int32), intent(in), optional :: obsId
    character(*),   intent(in), optional :: catTabName
    logical,        intent(in), optional :: useCatCoords

    type(SrcPrmT) :: catPrm
    ! integer(int32) :: obsIdNow
    logical :: useCatCoordsLoc
!     real(double), pointer, optional :: srcRaDataD(:), srcDecDataD(:)
! 
!     integer(int32), pointer, dimension(:) :: catObsIdDataInt
!     integer(int32), pointer :: catSrcNumData(:), catDetIdData(:)	! will be deallocated in release(catSet)
!     integer(int32), pointer :: catSrcIdDataF(:)
!     real(double),   pointer :: catRaDataD(:), catDecDataD(:)
!     type(ColumnT) :: catSrcNameCol

    !-------------------------------------------------------------------
    ! Start:

    useCatCoordsLoc = .false.	! Default
    if (present(useCatCoords)) then
      useCatCoordsLoc = useCatCoords
    end if

    ! Get parameters from the catalogue.
    call getCat5ColsObjFromSet(catSet, catPrm &
       , catTabName=catTabName, getRaDec=useCatCoordsLoc)
    call message(AppMsg, Noisy, 'Obtained Catalogue columns.')

    ! Then fill the srcPrm, using the above parameters.
    call set3To6ElementsInSrcPrm(catPrm, srcPrm, obsId=obsId)	! obsId if present()
    call message(AppMsg, Noisy, 'Catalogue data have been implemented.')

    if (verbosity(AppMsg) >= Noisy) then
       call message(AppMsg, Noisy, 'The catalogue source parameter for ID=1 is, ')
    end if

    call deallocateSrcPrm(catPrm &
       , iSrc=lbound(catPrm%obsId, 1), varName='catPrm', minVerbosity=Noisy)

  end subroutine set3To6ElmInSrcFromCatSet


  !---------------------------------------------------------------------
  subroutine getNew3ColsFromCatSet(catSet &
       , sumCatSrcNameData, sumSrcNumData, sumDetIdData, sumSrcIdData, sumInCatData &
       , obsId, catTabName, srcRaDataD, srcDecDataD)
  !---------------------------------------------------------------------
  ! Get three new columns (DETID, IAUNAME, INCATFLG) as arrays
  ! in the (summary) source list in the given obs.
  ! If SRCID does not exist, the values in sumSrcIdData are negative (-1).
  ! Make sure to deallocate all the pointers sum* and cat* after use.

    implicit none 

    type(DataSetT), intent(in) :: catSet
    character(*),   pointer :: sumCatSrcNameData(:)	! intent(inout), should be allocated beforehand.
    integer(int32), pointer :: sumSrcNumData(:), sumDetIdData(:), sumSrcIdData(:)	! intent(inout), should be allocated beforehand.
    logical(bool), pointer :: sumInCatData(:)				! intent(inout), should be allocated beforehand.
    integer(int32) :: obsId	! maybe obtained with getObsIdInt32FromHeaderFromSet(sumSet)
    character(*), intent(in), optional :: catTabName
    real(double), pointer, optional :: srcRaDataD(:), srcDecDataD(:)

    integer(int32), pointer, dimension(:) :: catObsIdDataInt
    integer(int32), pointer :: catSrcNumData(:), catDetIdData(:)	! will be deallocated in release(catSet)
    integer(int32), pointer :: catSrcIdDataF(:)
    real(double),   pointer :: catRaDataD(:), catDecDataD(:)
    type(ColumnT) :: catSrcNameCol

    if (present(srcRaDataD) .and. present(srcDecDataD)) then
      call getCat5ColsFromSet(catSet, catObsIdDataInt &
         , catSrcNumData, catDetIdData, catSrcIdDataF, catSrcNameCol &
         , catTabName=catTabName & 
         , catRaDataD=catRaDataD, catDecDataD=catDecDataD)	! catObsIdDataInt and catSrcIdDataF are allocated.
    else
      call getCat5ColsFromSet(catSet, catObsIdDataInt &
         , catSrcNumData, catDetIdData, catSrcIdDataF, catSrcNameCol &
         , catTabName=catTabName)	! catObsIdDataInt and catSrcIdDataF are allocated.
    end if

    call getNew3ColsFromCatColCatDataSId(catObsIdDataInt &
       , catSrcNumData, catDetIdData, catSrcIdDataF, catSrcNameCol &
       , sumCatSrcNameData, sumSrcNumData, sumDetIdData, sumSrcIdData, sumInCatData &
       , obsId)

    deallocate(catObsIdDataInt, catSrcIdDataF)
    if (present(srcRaDataD) .and. present(srcDecDataD)) then
      deallocate(catRaDataD, catDecDataD)
    end if

  end subroutine getNew3ColsFromCatSet


  !---------------------------------------------------------------------
  subroutine getNew3ColsFromColumnsSId(catObsIdDataInt &
       , catSrcNumData, catDetIdData, catSrcIdData, catSrcNameCol &
       , sumCatSrcNameCol, sumSrcNumData, sumDetIdData, sumSrcIdData, sumInCatData &
       , obsId)
  !---------------------------------------------------------------------
  ! Get three new columns in the (summary) source list in the given obs.
  ! Make sure to deallocate all the pointers cat* and sum*.

    implicit none 

    integer(int32), pointer, dimension(:) :: catObsIdDataInt	! intent(in)
    integer(int32), pointer :: catSrcNumData(:), catDetIdData(:), catSrcIdData(:)	! intent(in)
    type(ColumnT), intent(in) :: catSrcNameCol, sumCatSrcNameCol
    integer(int32), pointer :: sumSrcNumData(:)	! intent(in)
    integer(int32), pointer :: sumDetIdData(:), sumSrcIdData(:)	! intent(inout), should be allocated beforehand.
    logical(bool), pointer :: sumInCatData(:)				! intent(inout), should be allocated beforehand.
    integer(int32), intent(in) :: obsId

    !-------------------------------------------------------------------
    ! Start:

    call getNew3or4ColsFromColumns(catObsIdDataInt &
       , catSrcNumData, catDetIdData, catSrcNameCol &
       , sumCatSrcNameCol, sumSrcNumData, sumDetIdData, sumInCatData &
       , obsId, catSrcIdData, sumSrcIdData)

  end subroutine getNew3ColsFromColumnsSId


  !---------------------------------------------------------------------
  subroutine getNew3ColsFromColumnsNoS(catObsIdDataInt &
       , catSrcNumData, catDetIdData, catSrcIdData, catSrcNameCol &
       , sumCatSrcNameCol, sumSrcNumData, sumDetIdData, sumInCatData &
       , obsId)
  !---------------------------------------------------------------------
  ! Get three new columns in the (summary) source list in the given obs.
  ! Make sure to deallocate all the pointers cat* and sum*.

    implicit none 

    integer(int32), pointer, dimension(:) :: catObsIdDataInt	! intent(in)
    integer(int32), pointer :: catSrcNumData(:), catDetIdData(:), catSrcIdData(:)	! intent(in)
    type(ColumnT), intent(in) :: catSrcNameCol, sumCatSrcNameCol
    integer(int32), pointer :: sumSrcNumData(:)	! intent(in)
    integer(int32), pointer :: sumDetIdData(:)	! intent(inout), should be allocated beforehand.
    logical(bool), pointer :: sumInCatData(:)				! intent(inout), should be allocated beforehand.
    integer(int32), intent(in) :: obsId

    call getNew3or4ColsFromColumns(catObsIdDataInt &
       , catSrcNumData, catDetIdData, catSrcNameCol &
       , sumCatSrcNameCol, sumSrcNumData, sumDetIdData, sumInCatData &
       , obsId, catSrcIdData)	! catSrcIdData is not needed.

  end subroutine getNew3ColsFromColumnsNoS


  !---------------------------------------------------------------------
  subroutine getNew3or4ColsFromColumns(catObsIdDataInt &
       , catSrcNumData, catDetIdData, catSrcNameCol &
       , sumCatSrcNameCol, sumSrcNumData, sumDetIdData, sumInCatData &
       , obsId, catSrcIdData, sumSrcIdData)
  !---------------------------------------------------------------------
  ! Get three new columns in the (summary) source list in the given obs.
  ! Make sure to deallocate all the pointers cat* and sum*.

    implicit none 

    integer(int32), pointer, dimension(:) :: catObsIdDataInt	! intent(in)
    integer(int32), pointer :: catSrcNumData(:), catDetIdData(:)	! intent(in)
    type(ColumnT), intent(in) :: catSrcNameCol, sumCatSrcNameCol
    integer(int32), pointer :: sumSrcNumData(:)	! intent(in)
    integer(int32), pointer :: sumDetIdData(:)	! intent(inout), should be allocated beforehand.
    logical(bool), pointer :: sumInCatData(:)				! intent(inout), should be allocated beforehand.
    integer(int32), intent(in) :: obsId
    integer(int32), pointer, optional :: catSrcIdData(:)	! intent(in)
    integer(int32), pointer, optional :: sumSrcIdData(:)	! intent(inout), should be allocated beforehand.
    integer(int32) :: sumRow, catRow, numMatches
    character(LEN=10) :: srcNumStr	! LEN=3 must be enough.
    character(LENGTHOFIAUNAME) :: nameStr

    !-------------------------------------------------------------------
    ! Start:

    sumInCatData = .true. ! default
    do sumRow = lbound(sumSrcNumData, 1), ubound(sumSrcNumData, 1)
      numMatches = count(catObsIdDataInt==obsId&
           .and. catSrcNumData==sumSrcNumData(sumRow))

      if (numMatches > 1) then
        srcNumStr = intToStr(sumSrcNumData(sumRow))

        write(errstr, '(a, i10.10)') 'More than one entry was found &
             &in the catalog for source number '//trim(srcNumStr)//' in &
             &the summary source list of OBS-ID ', obsId
        call error('tooManyMatches')
      end if

      if (numMatches <= 0) then
        srcNumStr = intToStr(sumSrcNumData(sumRow))

! if (sumRow == 1) then
!    print *, 'obsId =', obsId
!    print *, 'catObsIdDataInt(1)=',catObsIdDataInt(1)
!    print *, 'catSrcNumData(1)=',catSrcNumData(1)
!    print *, 'sumSrcNumData(1)=',sumSrcNumData(1)
! end if
        write(errstr, '(a, i10.10)') 'No entry was found in the catalog &
             &for source number '//trim(srcNumStr)//' in the summary source &
             &list of OBS-ID ', obsId
        call warning('noMatchingSrc')
        call setStringCell(sumCatSrcNameCol, sumRow, BLANKSTRING)	! in DAL

        sumInCatData(sumRow) = .false.

      else
        catRow = smallestTrueIndex(catObsIdDataInt==obsId&
           .and. catSrcNumData==sumSrcNumData(sumRow)) - 1

        sumDetIdData(sumRow) = catDetIdData(catRow)	! Added DETID
        if (present(catSrcIdData) .and. present(sumSrcIdData)) then
          sumSrcIdData(sumRow) = catSrcIdData(catRow)	! Added SRCID
        end if

        nameStr = '' ! prevents funny characters at the end if the string
        ! in catSrcNameCol is shorter than nameStr.

        nameStr = stringCell(catSrcNameCol, catRow)
        call setStringCell(sumCatSrcNameCol, sumRow, nameStr)	! in DAL
!print *, 'sumRow, catRow, sumSrcNumData(sumRow), catSrcNumData(catRow): '&
!, sumRow, catRow, sumSrcNumData(sumRow), catSrcNumData(catRow)
!print *, 'nameStr: ', trim(nameStr)
      end if
    end do	! do sumRow = lbound(sumSrcNumData, 1), ubound(sumSrcNumData, 1)

  end subroutine getNew3or4ColsFromColumns


!   !---------------------------------------------------------------------
!   subroutine getNew3ColsFromColumns(catObsIdDataInt &
!        , catSrcNumData, catDetIdData, catSrcIdData, catSrcNameCol &
!        , sumCatSrcNameCol, sumSrcNumData, sumDetIdData, sumSrcIdData, sumInCatData &
!        , obsId)
!   !---------------------------------------------------------------------
!   ! Get three new columns in the (summary) source list in the given obs.
!   ! Make sure to deallocate all the pointers cat* and sum*.
! 
!     implicit none 
! 
!     integer(int32), pointer, dimension(:) :: catObsIdDataInt	! intent(in)
!     integer(int32), pointer :: catSrcNumData(:), catDetIdData(:), catSrcIdData(:)	! intent(in)
!     type(ColumnT), intent(in) :: catSrcNameCol, sumCatSrcNameCol
!     integer(int32), pointer :: sumSrcNumData(:)	! intent(in)
!     integer(int32), pointer :: sumDetIdData(:), sumSrcIdData(:)	! intent(inout), should be allocated beforehand.
!     logical(bool), pointer :: sumInCatData(:)				! intent(inout), should be allocated beforehand.
!     integer(int32), intent(in) :: obsId
!     integer(int32) :: sumRow, catRow, numMatches
!     character(LEN=10) :: srcNumStr
!     character(LENGTHOFIAUNAME) :: nameStr
! 
!     !-------------------------------------------------------------------
!     ! Start:
! 
!     sumInCatData = .true. ! default
!     do sumRow = lbound(sumSrcNumData, 1), ubound(sumSrcNumData, 1)
!       numMatches = count(catObsIdDataInt==obsId&
!            .and. catSrcNumData==sumSrcNumData(sumRow))
! 
!       if (numMatches > 1) then
!         srcNumStr = intToStr(sumSrcNumData(sumRow))
! 
!         write(errstr, '(a, i10.10)') 'More than one entry was found &
!              &in the catalog for source number '//trim(srcNumStr)//' in &
!              &the summary source list of OBS-ID ', obsId
!         call error('tooManyMatches')
!       end if
! 
!       if (numMatches <= 0) then
!         srcNumStr = intToStr(sumSrcNumData(sumRow))
! 
!         write(errstr, '(a, i10.10)') 'No entry was found in the catalog &
!              &for source number '//trim(srcNumStr)//' in the summary source &
!              &list of OBS-ID ', obsId
!         call warning('noMatchingSrc')
!         call setStringCell(sumCatSrcNameCol, sumRow, BLANKSTRING)	! in DAL
! 
!         sumInCatData(sumRow) = .false.
! 
!       else
!         catRow = smallestTrueIndex(catObsIdDataInt==obsId&
!            .and. catSrcNumData==sumSrcNumData(sumRow)) - 1
! 
!         sumDetIdData(sumRow) = catDetIdData(catRow)	! Added DETID
!         sumSrcIdData(sumRow) = catSrcIdData(catRow)	! Added SRCID
! 
!         nameStr = '' ! prevents funny characters at the end if the string
!         ! in catSrcNameCol is shorter than nameStr.
! 
!         nameStr = stringCell(catSrcNameCol, catRow)
!         call setStringCell(sumCatSrcNameCol, sumRow, nameStr)	! in DAL
! !print *, 'sumRow, catRow, sumSrcNumData(sumRow), catSrcNumData(catRow): '&
! !, sumRow, catRow, sumSrcNumData(sumRow), catSrcNumData(catRow)
! !print *, 'nameStr: ', trim(nameStr)
!       end if
!     end do	! do sumRow = lbound(sumSrcNumData, 1), ubound(sumSrcNumData, 1)
! 
!   end subroutine getNew3ColsFromColumns


  !---------------------------------------------------------------------
  subroutine getNew3ColsFromCatColCatDataSId(catObsIdDataInt &
       , catSrcNumData, catDetIdData, catSrcIdData, catSrcNameCol &
       , sumCatSrcNameData, sumSrcNumData, sumDetIdData, sumSrcIdData, sumInCatData &
       , obsId)
  !---------------------------------------------------------------------
  ! Get three new columns in the (summary) source list in the given obs.
  ! With SrcId.
  ! Make sure to deallocate all the pointers cat* and sum*.

    implicit none 

    integer(int32), pointer, dimension(:) :: catObsIdDataInt	! intent(in)
    integer(int32), pointer :: catSrcNumData(:), catDetIdData(:), catSrcIdData(:)	! intent(in)
    type(ColumnT), intent(in) :: catSrcNameCol
    character(*),   pointer :: sumCatSrcNameData(:)	! intent(inout), should be allocated beforehand.
    integer(int32), pointer :: sumSrcNumData(:)	! intent(in)
    integer(int32), pointer :: sumDetIdData(:), sumSrcIdData(:)	! intent(inout), should be allocated beforehand.
    logical(bool), pointer :: sumInCatData(:)				! intent(inout), should be allocated beforehand.
    integer(int32), intent(in) :: obsId

    call getNew3Or4ColsFromCatColCatData(catObsIdDataInt &
       , catSrcNumData, catDetIdData, catSrcNameCol &
       , sumCatSrcNameData, sumSrcNumData, sumDetIdData, sumInCatData &
       , obsId, catSrcIdData, sumSrcIdData)	! nb. Change of order of ARGs

  end subroutine getNew3ColsFromCatColCatDataSId


  !---------------------------------------------------------------------
  subroutine getNew3ColsFromCatColCatDataNoS(catObsIdDataInt &
       , catSrcNumData, catDetIdData, catSrcIdData, catSrcNameCol &
       , sumCatSrcNameData, sumSrcNumData, sumDetIdData, sumInCatData &
       , obsId)
  !---------------------------------------------------------------------
  ! Get three new columns in the (summary) source list in the given obs.
  ! No SrcId, though.
  ! Make sure to deallocate all the pointers cat* and sum*.

    implicit none 

    integer(int32), pointer, dimension(:) :: catObsIdDataInt	! intent(in)
    integer(int32), pointer :: catSrcNumData(:), catDetIdData(:), catSrcIdData(:)	! intent(in)
    type(ColumnT), intent(in) :: catSrcNameCol
    character(*),   pointer :: sumCatSrcNameData(:)	! intent(inout), should be allocated beforehand.
    integer(int32), pointer :: sumSrcNumData(:)	! intent(in)
    integer(int32), pointer :: sumDetIdData(:)	! intent(inout), should be allocated beforehand.
    logical(bool), pointer :: sumInCatData(:)				! intent(inout), should be allocated beforehand.
    integer(int32), intent(in) :: obsId

    call getNew3Or4ColsFromCatColCatData(catObsIdDataInt &
       , catSrcNumData, catDetIdData, catSrcNameCol &
       , sumCatSrcNameData, sumSrcNumData, sumDetIdData, sumInCatData &
       , obsId, catSrcIdData)	! nb. Change of order of ARGs

  end subroutine getNew3ColsFromCatColCatDataNoS


  !---------------------------------------------------------------------
  subroutine set3To6ElementsInSrcPrm(catPrm, srcPrm, obsId)
  !---------------------------------------------------------------------
  ! Set columns of DETID, IAUNAME, INCATFLG, and optionally
  ! SRCID, RA, DEC in the (summary) source list (SrcPrmT object),
  ! using the external catalogue (catPrm).
  ! Almost the same as getNew3Or4ColsFromCatColCatData() but
  ! with different arguments (and supports for RA and DEC).
  ! Make sure to call deallocateSrcPrm(catPrm, srcPrm) after use.

    implicit none 

    type(SrcPrmT), intent(inout) :: catPrm, srcPrm
    integer(int32), intent(in), optional :: obsId

    integer(int32) :: obsIdNow
    integer(int32) :: sumRow, catRow, numMatches, lboundCatPrm, lboundCatPrmCorr
    character(LEN=10) :: srcNumStr	! LEN=3 must be enough.
    character(LENGTHOFIAUNAME) :: nameStr

    !-------------------------------------------------------------------
    ! Start:

    ! Sanity checks
    if (.not. srcPrm%flagAlloc%srcNum)  call error ('Alloc_srcNum_getNew3Or4Obj1','')
    !if (.not. srcPrm%flagAlloc%detId)   call error ('Alloc_detId_getNew3Or4Obj1','')
    if (.not. srcPrm%flagAlloc%catName) call error ('Alloc_catName_getNew3Or4Obj1','')
    if (.not. srcPrm%flagAlloc%inCatFlag) call error ('Alloc_inCatFlag_getNew3Or4Obj1','')

    if (.not. catPrm%flagAlloc%obsId)   call error ('Alloc_obsId_getNew3Or4Obj2','')
    if (.not. catPrm%flagAlloc%srcNum)  call error ('Alloc_srcNum_getNew3Or4Obj2','')
    !if (.not. catPrm%flagAlloc%detId)   call error ('Alloc_detId_getNew3Or4Obj2','')

    if (present(obsId)) then
      obsIdNow = obsId
    else
      if (.not. srcPrm%flagAlloc%obsId)   call error ('Alloc_obsId_getNew3Or4Obj1','')
      obsIdNow = srcPrm%obsId(1)
    end if

    ! ----- Main -----

    lboundCatPrm = lbound(catPrm%obsId, 1)
    lboundCatPrmCorr = 1 - lboundCatPrm	! NOTE: A positive index X
    ! (Fortran standard) is translated into the true index
    ! of X-lboundCatPrmCorr;
    !  e.g., Array(1:):  X => X   (Fortran-Standard)
    !      , Array(0:):  X => X-1 (C-style)
    !      , Array(-3:): X => X-4

    srcPrm%inCatFlag = .true. ! default INCATFLG

    do sumRow = lbound(srcPrm%srcNum, 1), ubound(srcPrm%srcNum, 1)
      numMatches = count(catPrm%obsId==obsIdNow&
           .and. catPrm%srcNum==srcPrm%srcNum(sumRow))

      if (numMatches > 1) then
        srcNumStr = intToStr(srcPrm%srcNum(sumRow))
        write(errstr, '(a, i10.10)') 'More than one entry was found &
             &in the catalog for source number '//trim(srcNumStr)//' in &
             &the summary source list of OBS-ID ', obsIdNow
        call error('tooManyMatches')

      else if (numMatches <= 0) then
        srcNumStr = intToStr(srcPrm%srcNum(sumRow))
        write(errstr, '(a, i10.10)') 'No entry was found in the catalog &
             &for source number '//trim(srcNumStr)//' in the summary source &
             &list of OBS-ID ', obsIdNow
        call warning('noMatchingSrc')

        srcPrm%catName(sumRow) = trim(BLANKSTRING)
        srcPrm%inCatFlag(sumRow) = .false.

      else
        catRow = smallestTrueIndex(catPrm%obsId==obsIdNow&
           .and. catPrm%srcNum==srcPrm%srcNum(sumRow))
        catRow = catRow - lboundCatPrmCorr

        write(errstr, '(a, i6, a, i8, a, i2, a)') &
           "SRC_NUM=(", srcPrm%srcNum(sumRow) &
           , ") matches the external catalogue Row=(", catRow &
           , ") (Start=", lboundCatPrm, ")"
        call message(AppMsg, Noisy)

        if (catPrm%flagAlloc%detid  .and. srcPrm%flagAlloc%detid) then
          srcPrm%detid(sumRow) = catPrm%detid(catRow)	! Added DETID
        end if
        if (catPrm%flagAlloc%srcid  .and. srcPrm%flagAlloc%srcid) then
          srcPrm%srcid(sumRow)  = catPrm%srcid(catRow)	! Added SRCID
        end if
        if (catPrm%flagAlloc%raDeg  .and. srcPrm%flagAlloc%raDeg) then
          srcPrm%raDeg(sumRow)  = catPrm%raDeg(catRow)	! Added RA
          ! NOTE: catPrm%raDeg is not always allocated.
        end if
        if (catPrm%flagAlloc%decDeg .and. srcPrm%flagAlloc%decDeg) then
          srcPrm%decDeg(sumRow) = catPrm%decDeg(catRow)	! Added DEC
        end if

        nameStr = '' ! prevents funny characters at the end if the string
          ! in catPrm%catNameCol is shorter than nameStr.
        nameStr = stringCell(catPrm%catNameCol, catRow)
        srcPrm%catName(sumRow) = trim(nameStr)		! Added IAUNAME
      end if
    end do	! do sumRow = lbound(sumSrcNumData, 1), ubound(sumSrcNumData, 1)

  end subroutine set3To6ElementsInSrcPrm


  !---------------------------------------------------------------------
  subroutine getNew3Or4ColsFromCatColCatData(catObsIdDataInt &
       , catSrcNumData, catDetIdData, catSrcNameCol &
       , sumCatSrcNameData, sumSrcNumData, sumDetIdData, sumInCatData &
       , obsId, catSrcIdData, sumSrcIdData)
  !---------------------------------------------------------------------
  ! Get three new columns in the (summary) source list in the given obs.
  ! Make sure to deallocate all the pointers cat* and sum*.
  ! See also set3To6ElementsInSrcPrm(), which is very similar
  ! (but with the additional support of RA and DEC).

    implicit none 

    integer(int32), pointer, dimension(:) :: catObsIdDataInt	! intent(in)
    integer(int32), pointer :: catSrcNumData(:), catDetIdData(:)	! intent(in)
    type(ColumnT), intent(in) :: catSrcNameCol
    character(*),   pointer :: sumCatSrcNameData(:)	! intent(inout), should be allocated beforehand.
    integer(int32), pointer :: sumSrcNumData(:)	! intent(in)
    integer(int32), pointer :: sumDetIdData(:)	! intent(inout), should be allocated beforehand.
    logical(bool), pointer :: sumInCatData(:)				! intent(inout), should be allocated beforehand.
    integer(int32), intent(in) :: obsId
    integer(int32), pointer, optional :: catSrcIdData(:)	! intent(in)
    integer(int32), pointer, optional :: sumSrcIdData(:)	! intent(inout), should be allocated beforehand.

    integer(int32) :: sumRow, catRow, numMatches
    character(LEN=10) :: srcNumStr	! LEN=3 must be enough.
    character(LENGTHOFIAUNAME) :: nameStr

    !-------------------------------------------------------------------
    ! Start:

    sumInCatData = .true. ! default
    do sumRow = lbound(sumSrcNumData, 1), ubound(sumSrcNumData, 1)
      numMatches = count(catObsIdDataInt==obsId&
           .and. catSrcNumData==sumSrcNumData(sumRow))

      if (numMatches > 1) then
        srcNumStr = intToStr(sumSrcNumData(sumRow))

        write(errstr, '(a, i10.10)') 'More than one entry was found &
             &in the catalog for source number '//trim(srcNumStr)//' in &
             &the summary source list of OBS-ID ', obsId
        call error('tooManyMatches')
      end if

      if (numMatches <= 0) then
        srcNumStr = intToStr(sumSrcNumData(sumRow))

        write(errstr, '(a, i10.10)') 'No entry was found in the catalog &
             &for source number '//trim(srcNumStr)//' in the summary source &
             &list of OBS-ID ', obsId
        call warning('noMatchingSrc')
        sumCatSrcNameData(sumRow) = trim(BLANKSTRING)

        sumInCatData(sumRow) = .false.

      else
        catRow = smallestTrueIndex(catObsIdDataInt==obsId&
           .and. catSrcNumData==sumSrcNumData(sumRow)) - 1

        sumDetIdData(sumRow) = catDetIdData(catRow)	! Added DETID
        if (present(catSrcIdData) .and. present(sumSrcIdData)) then
          sumSrcIdData(sumRow) = catSrcIdData(catRow)	! Added SRCID
        end if

        nameStr = '' ! prevents funny characters at the end if the string
        ! in catSrcNameCol is shorter than nameStr.

        nameStr = stringCell(catSrcNameCol, catRow)
        sumCatSrcNameData(sumRow) = trim(nameStr)
!print *, 'sumRow, catRow, sumSrcNumData(sumRow), catSrcNumData(catRow): '&
!, sumRow, catRow, sumSrcNumData(sumRow), catSrcNumData(catRow)
!print *, 'nameStr: ', trim(nameStr)
      end if
    end do	! do sumRow = lbound(sumSrcNumData, 1), ubound(sumSrcNumData, 1)

  end subroutine getNew3Or4ColsFromCatColCatData


!   !---------------------------------------------------------------------
!   subroutine getNew3ColsFromCatColCatData(catObsIdDataInt &
!        , catSrcNumData, catDetIdData, catSrcIdData, catSrcNameCol &
!        , sumCatSrcNameData, sumSrcNumData, sumDetIdData, sumSrcIdData, sumInCatData &
!        , obsId)
!   !---------------------------------------------------------------------
!   ! Get three new columns in the (summary) source list in the given obs.
!   ! Make sure to deallocate all the pointers cat* and sum*.
! 
!     implicit none 
! 
!     integer(int32), pointer, dimension(:) :: catObsIdDataInt	! intent(in)
!     integer(int32), pointer :: catSrcNumData(:), catDetIdData(:), catSrcIdData(:)	! intent(in)
!     type(ColumnT), intent(in) :: catSrcNameCol
!     character(*),   pointer :: sumCatSrcNameData(:)	! intent(inout), should be allocated beforehand.
!     integer(int32), pointer :: sumSrcNumData(:)	! intent(in)
!     integer(int32), pointer :: sumDetIdData(:), sumSrcIdData(:)	! intent(inout), should be allocated beforehand.
!     logical(bool), pointer :: sumInCatData(:)				! intent(inout), should be allocated beforehand.
!     integer(int32), intent(in) :: obsId
!     integer(int32) :: sumRow, catRow, numMatches
!     character(LEN=10) :: srcNumStr
!     character(LENGTHOFIAUNAME) :: nameStr
! 
!     !-------------------------------------------------------------------
!     ! Start:
! 
!     sumInCatData = .true. ! default
!     do sumRow = lbound(sumSrcNumData, 1), ubound(sumSrcNumData, 1)
!       numMatches = count(catObsIdDataInt==obsId&
!            .and. catSrcNumData==sumSrcNumData(sumRow))
! 
!       if (numMatches > 1) then
!         srcNumStr = intToStr(sumSrcNumData(sumRow))
! 
!         write(errstr, '(a, i10.10)') 'More than one entry was found &
!              &in the catalog for source number '//trim(srcNumStr)//' in &
!              &the summary source list of OBS-ID ', obsId
!         call error('tooManyMatches')
!       end if
! 
!       if (numMatches <= 0) then
!         srcNumStr = intToStr(sumSrcNumData(sumRow))
! 
!         write(errstr, '(a, i10.10)') 'No entry was found in the catalog &
!              &for source number '//trim(srcNumStr)//' in the summary source &
!              &list of OBS-ID ', obsId
!         call warning('noMatchingSrc')
!         sumCatSrcNameData(sumRow) = trim(BLANKSTRING)
! 
!         sumInCatData(sumRow) = .false.
! 
!       else
!         catRow = smallestTrueIndex(catObsIdDataInt==obsId&
!            .and. catSrcNumData==sumSrcNumData(sumRow)) - 1
! 
!         sumDetIdData(sumRow) = catDetIdData(catRow)	! Added DETID
!         sumSrcIdData(sumRow) = catSrcIdData(catRow)	! Added SRCID
! 
!         nameStr = '' ! prevents funny characters at the end if the string
!         ! in catSrcNameCol is shorter than nameStr.
! 
!         nameStr = stringCell(catSrcNameCol, catRow)
!         sumCatSrcNameData(sumRow) = trim(nameStr)
! !print *, 'sumRow, catRow, sumSrcNumData(sumRow), catSrcNumData(catRow): '&
! !, sumRow, catRow, sumSrcNumData(sumRow), catSrcNumData(catRow)
! !print *, 'nameStr: ', trim(nameStr)
!       end if
!     end do	! do sumRow = lbound(sumSrcNumData, 1), ubound(sumSrcNumData, 1)
! 
!   end subroutine getNew3ColsFromCatColCatData


  !---------------------------------------------------------------------
  function getObsIdInt32FromHeaderFromName(filename) result(obsId)
  !---------------------------------------------------------------------
  ! Get OBS-ID as int32 from the FITS header.

    implicit none 

    integer(int32) :: obsId
    character(*), intent(in) :: filename

    type(DataSetT) :: inSet

    !-------------------------------------------------------------------
    ! Start:

    inSet = dataSet(filename, READ)
    obsId = getObsIdInt32FromHeaderFromSet(inSet)
    call release(inSet)

  end function getObsIdInt32FromHeaderFromName


  !---------------------------------------------------------------------
  function getObsIdInt32FromHeaderFromSet(inSet) result(obsId)
  !---------------------------------------------------------------------
  ! Get OBS-ID as int32 from the FITS header

    implicit none 

    integer(int32) :: obsId
    type(DataSetT), intent(in) :: inSet

    character(LENOBSIDSTR), parameter :: arKeyObsIdStr(4) &
       = (/'OBS_ID', 'obs_id', 'OBSID ', 'obsid ' /)
    character(LEN=LENOBSIDSTR) :: keyObsIdStr
    character(LEN=LENSTRATTR) :: tmpStr
    logical :: isString &		! Default: .true.
       , isObsIdFound
    integer(int32) :: numBlock
    integer :: i

    !-------------------------------------------------------------------
    ! Start:

    isString = .true.
    isObsIdFound = .false.	! unnecessary, but just to play safe.
    do i=lbound(arKeyObsIdStr, 1), ubound(arKeyObsIdStr, 1) 
      keyObsIdStr = trim(arKeyObsIdStr(i))
      if (trim(keyObsIdStr) == 'OBSID') isString = .false.

      if (isString) then 
        call retKwdFromAnyBlock(tmpStr, inSet, trim(keyObsIdStr) &
           , numBlock, isObsIdFound)
      else
        call retKwdFromAnyBlock(obsId,  inSet, trim(keyObsIdStr) &
           , numBlock, isObsIdFound)
      end if

      if (isObsIdFound) then
        if (isString) then
          ! read(trim(tmpStr), *) obsId	! Causes a compile error.
          tmpStr = trim(tmpStr)		! To play safe
          read(tmpStr, *) obsId
        end if
        if (numBlock .ne. -1) then	! Not in the Primary header
          write(errStr, '(a, i3, a)') &
             'OBS_ID (or alike) is not found in the primary header &
             &but in the header of the extention No.', numBlock, &
             ' where the first extention is counted as 0-th (not 1st).'
          call warning('NotFoundObsIdInPrimary')
        end if

        return	! ObsId is found.
      end if

      ! if (hasAttribute(inSet, keyObsIdStr)) then
      !   isObsIdFound = .true.
      !   exit
      ! end if
    end do

    !! If the code reaches here, it raises Error.
    call release(inSet)
    call error('NotFoundObsIdInHeader', 'OBS_ID (or alike) not found.')

    ! ! if (hasAttribute(inSet, 'OBS_ID')) then
    ! !   keyObsIdStr = 'OBS_ID'
    ! ! else if (hasAttribute(inSet, 'obs_id')) then
    ! !   keyObsIdStr = 'obs_id'
    ! ! else if (hasAttribute(inSet, 'OBSID')) then
    ! !   keyObsIdStr = 'OBSID'
    ! !   isString = .false.
    ! ! else if (hasAttribute(inSet, 'obsid')) then
    ! !   keyObsIdStr = 'obsid'
    ! !   isString = .false.
    ! ! else
    ! 
    ! if (.not. isObsIdFound) then
    !   call release(inSet)
    !   call error('NotFoundObsIdInHeader', 'OBS_ID (or alike) not found.')
    ! end if
    ! 
    ! if (isString) then
    !    tmpStr = trim(stringAttribute(inSet, trim(keyObsIdStr)))
    !    read(tmpStr, *) obsId
    ! else
    !    obsId = int32Attribute(inSet, trim(keyObsIdStr))
    ! end if

  end function getObsIdInt32FromHeaderFromSet


  !---------------------------------------------------------------------
  function getObsIdInt32FromHeaderFromTab(inTab) result(obsId)
  !---------------------------------------------------------------------
  ! Get OBS-ID as int32 from the FITS header

    implicit none 

    integer(int32) :: obsId
    type(TableT), intent(in) :: inTab

    character(LEN=LENOBSIDSTR) :: keyObsIdStr
    character(LEN=LENSTRATTR) :: tmpStr
    logical :: isString		! Default: .true.

    !-------------------------------------------------------------------
    ! Start:

    isString = .true.
    if (hasAttribute(inTab, 'OBS_ID')) then
      keyObsIdStr = 'OBS_ID'
    else if (hasAttribute(inTab, 'obs_id')) then
      keyObsIdStr = 'obs_id'
    else if (hasAttribute(inTab, 'OBSID')) then
      keyObsIdStr = 'OBSID'
      isString = .false.
    else if (hasAttribute(inTab, 'obsid')) then
      keyObsIdStr = 'obsid'
      isString = .false.
    else
      call error('NotFoundObsIdInHeader', 'OBS_ID (or alike) not found.')
    end if

    if (isString) then
       tmpStr = trim(stringAttribute(inTab, trim(keyObsIdStr)))
       read(tmpStr, *) obsId
    else
       obsId = int32Attribute(inTab, trim(keyObsIdStr))
    end if

  end function getObsIdInt32FromHeaderFromTab


  !---------------------------------------------------------------------
  function getObsIdStr(catTab, keyw, noErr) result(retStr)
  !---------------------------------------------------------------------
  ! Return the column title name of OBSERVATION ID (such as OBS_ID).

    implicit none 

    character(LEN=LENOBSIDSTR) :: retStr
    type(TableT), intent(in) :: catTab
    character(*), intent(in), optional :: keyw
    logical, intent(in), optional :: noErr	! Default: .false.

    character(LEN=LENSTRCOLNAME), parameter :: arKeyw(4) &
       = (/ 'OBSID ', 'obsid ', 'OBS_ID', 'obs_id' /)
    character(LEN=LENSTRCOLNAME), parameter :: errNameNow = 'NotFoundOBSID'

    if (present(keyw)) then
      retStr = getColKwdStrFromKey(catTab, keyw &
         , errName=trim(errNameNow), noErr=noErr)
    else
      retStr = getColKwdStrFromAry(catTab, arKeyw &
         , errName=trim(errNameNow), noErr=noErr)
    end if
 
  end function getObsIdStr


  !---------------------------------------------------------------------
  function getSrcNumStr(catTab, keyw, noErr) result(retStr)
  !---------------------------------------------------------------------
  ! Return the column title name of SRC_NUM.
  ! If noErr option is specified and is true, then this does not issue error,
  ! and returns '' if an erroneous situation happens.

    implicit none 

    character(LEN=LENSTRCOLNAME) :: retStr
    type(TableT), intent(in) :: catTab
    character(*), intent(in), optional :: keyw
    logical, intent(in), optional :: noErr	! Default: .false.

    character(LEN=LENSTRCOLNAME), parameter :: arKeyw(4) &
       = (/ 'SRC_NUM    ', 'src_num    ', 'SUM_SRC_NUM', 'sum_src_num' /)
    character(LEN=LENSTRCOLNAME), parameter :: errNameNow = 'NotFoundSRCNUM'

    if (present(keyw)) then
      retStr = getColKwdStrFromKey(catTab, keyw &
         , errName=trim(errNameNow), noErr=noErr)
    else
      retStr = getColKwdStrFromAry(catTab, arKeyw &
         , errName=trim(errNameNow), noErr=noErr)
    end if
 
  end function getSrcNumStr


  !---------------------------------------------------------------------
  function getIAUNameStr(catTab, keyw, noErr) result(retStr)
  !---------------------------------------------------------------------
  ! Return the column title name of IAUNAME.
  ! If noErr option is specified and is true, then this does not issue error,
  ! and returns '' if an erroneous situation happens.

    implicit none 

    character(LEN=LENSTRCOLNAME) :: retStr
    type(TableT), intent(in) :: catTab
    character(*), intent(in), optional :: keyw
    logical, intent(in), optional :: noErr	! Default: .false.

    character(LEN=LENSTRCOLNAME), parameter :: arKeyw(4) &
       = (/ 'IAUNAME   ', 'iauname   ', 'XMMSRCNAME', 'xmmsrcname' /)
    character(LEN=LENSTRCOLNAME), parameter :: errNameNow = 'NotFoundIAUNAME'

    if (present(keyw)) then
      retStr = getColKwdStrFromKey(catTab, keyw &
         , errName=trim(errNameNow), noErr=noErr)
    else
      retStr = getColKwdStrFromAry(catTab, arKeyw &
         , errName=trim(errNameNow), noErr=noErr)
    end if
 
  end function getIAUNameStr


  !---------------------------------------------------------------------
  function getDetIdStr(catTab, keyw, noErr) result(retStr)
  !---------------------------------------------------------------------
  ! Return the column title name of DETID.
  ! If noErr option is specified and is true, then this does not issue error,
  ! and returns '' if an erroneous situation happens.

    implicit none 

    character(LEN=LENSTRCOLNAME) :: retStr
    type(TableT), intent(in) :: catTab
    character(*), intent(in), optional :: keyw
    logical, intent(in), optional :: noErr	! Default: .false.

    character(LEN=LENSTRCOLNAME), parameter :: arKeyw(4) &
       = (/ 'DETID ', 'detid ', 'DET_ID', 'det_id' /)
    character(LEN=LENSTRCOLNAME), parameter :: errNameNow = 'NotFoundDETID'

    if (present(keyw)) then
      retStr = getColKwdStrFromKey(catTab, keyw &
         , errName=trim(errNameNow), noErr=noErr)
    else
      retStr = getColKwdStrFromAry(catTab, arKeyw &
         , errName=trim(errNameNow), noErr=noErr)
    end if
 
  end function getDetIdStr

  !---------------------------------------------------------------------
  function getSrcIdStr(catTab, keyw, noErr) result(retStr)
  !---------------------------------------------------------------------
  ! Return the column title name of SRCID.
  ! If noErr option is specified and is true, then this does not issue error,
  ! and returns '' if an erroneous situation happens.

    implicit none 

    character(LEN=LENSTRCOLNAME) :: retStr
    type(TableT), intent(in) :: catTab
    character(*), intent(in), optional :: keyw
    logical, intent(in), optional :: noErr	! Default: .false.

    character(LEN=LENSTRCOLNAME), parameter :: arKeyw(4) &
       = (/ 'SRCID ', 'srcid ', 'SRC_ID', 'src_id' /)
    character(LEN=LENSTRCOLNAME), parameter :: errNameNow = 'NotFoundSRCID'

    if (present(keyw)) then
      retStr = getColKwdStrFromKey(catTab, keyw &
         , errName=trim(errNameNow), noErr=noErr)
    else
      retStr = getColKwdStrFromAry(catTab, arKeyw &
         , errName=trim(errNameNow), noErr=noErr)
    end if
 
  end function getSrcIdStr

  !---------------------------------------------------------------------
  function getInCatFlgStr(catTab, keyw, noErr) result(retStr)
  !---------------------------------------------------------------------
  ! Return the column title name of INCATFLG.
  ! If noErr option is specified and is true (or simply not specified), then this does not issue error,
  ! and returns '' if an erroneous situation happens.

    implicit none 

    character(LEN=LENSTRCOLNAME) :: retStr
    type(TableT), intent(in) :: catTab
    character(*), intent(in), optional :: keyw
    logical, intent(in), optional :: noErr	! Default: .true.

    character(LEN=LENSTRCOLNAME), parameter :: arKeyw(2) &
       = (/ 'INCATFLG', 'incatflg' /)
    character(LEN=LENSTRCOLNAME), parameter :: errNameNow = 'NotFoundINCATFLG'

    if (present(keyw)) then
      retStr = getColKwdStrFromKey(catTab, keyw &
         , errName=trim(errNameNow), noErr=noErr)
    else
      retStr = getColKwdStrFromAry(catTab, arKeyw &
         , errName=trim(errNameNow), noErr=noErr)
    end if
 
  end function getInCatFlgStr

!   !---------------------------------------------------------------------
!   function getObsIdStr(catTab)
!   !---------------------------------------------------------------------
!   ! Return the column title name of OBSERVATION ID (such as OBS_ID).
! 
!     implicit none 
! 
!     character(LEN=LENOBSIDSTR) :: getObsIdStr
!     type(TableT), intent(in) :: catTab
! 
!     if (hasColumn(catTab, 'OBSID')) then
!       getObsIdStr = 'OBSID'
!     else if (hasColumn(catTab, 'OBS_ID')) then
!       getObsIdStr = 'OBS_ID'
!     else if (hasColumn(catTab, 'obsid')) then
!       getObsIdStr = 'obsid'
!     else if (hasColumn(catTab, 'obs_id')) then
!       getObsIdStr = 'obs_id'
!     else
!       call error('NotFoundObsId', 'OBSID (or alike) not found.')
!       ! call error('NotFoundObsId', 'OBSID not found in the input Summary catalogue.')
!     end if
! 
!   end function getObsIdStr
! 
! 
!   !---------------------------------------------------------------------
!   function getSrcNumStr(catTab, keyw) result(retStr)
!   !---------------------------------------------------------------------
!   ! Return the column title name of SRC_NUM.
! 
!     implicit none 
! 
!     character(LEN=LENOBSIDSTR) :: retStr
!     type(TableT), intent(in) :: catTab
!     character(*), intent(in), optional :: keyw
! 
!     if (present(keyw)) then
!       if (hasColumn(catTab, trim(keyw))) then
!         retStr = trim(keyw)
!       else
!         call error('NotFoundSrcNum', 'Keyword '//trim(keyw)//' not found.')
!       end if
!     else
!       if (hasColumn(catTab, 'SRC_NUM')) then
!         retStr = 'SRC_NUM'
!       else if (hasColumn(catTab, 'SUM_SRC_NUM')) then
!         retStr = 'SUM_SRC_NUM'
!       else if (hasColumn(catTab, 'src_num')) then
!         retStr = 'src_num'
!       else
!         call error('NotFoundSrcNum', 'SRC_NUM (or alike) not found.')
!         !call error('NotFoundSrcNum', 'SRC_NUM not found in the input catalogue.')
!       end if
!     end if
! 
!   end function getSrcNumStr
! 
! 
!   !---------------------------------------------------------------------
!   function getIAUNameStr(catTab, keyw) result(retStr)
!   !---------------------------------------------------------------------
!   ! Return the column title name of IAUNAME.
! 
!     implicit none 
! 
!     character(LEN=LENOBSIDSTR) :: retStr
!     type(TableT), intent(in) :: catTab
!     character(*), intent(in), optional :: keyw
! 
!     if (present(keyw)) then
!       if (hasColumn(catTab, trim(keyw))) then
!         retStr = trim(keyw)
!       else
!         call error('NotFoundIAUNAME', 'Keyword '//trim(keyw)//' not found.')
!       end if
!     else
!       if (hasColumn(catTab, 'IAUNAME')) then
!         retStr = 'IAUNAME'
!       else if (hasColumn(catTab, 'XMMSRCNAME')) then
!         retStr = 'XMMSRCNAME'
!       else if (hasColumn(catTab, 'iauname')) then
!         retStr = 'iauname'
!       else
!         call error('NotFoundIAUNAME', 'IAUNAME (or alike) not found.')
!       end if
!     end if
! 
!   end function getIAUNameStr
! 
! 
!   !---------------------------------------------------------------------
!   function getDetIdStr(catTab, keyw, noErr) result(retStr)
!   !---------------------------------------------------------------------
!   ! Return the column title name of DETID.
!   ! If noErr option is specified and is true, then this does not issue error,
!   ! and returns '' if an erroneous situation happens.
! 
!     implicit none 
! 
!     character(LEN=LENOBSIDSTR) :: retStr
!     type(TableT), intent(in) :: catTab
!     character(*), intent(in), optional :: keyw
!     logical, intent(in), optional :: noErr	! Default: .false.
! 
!     if (present(keyw)) then
!       if (hasColumn(catTab, trim(keyw))) then
!         retStr = trim(keyw)
!       else
!         if (present(noErr)) then
!           if (noErr) then
!             retStr = ''
!           else
!             call error('NotFoundDETID', 'Keyword '//trim(keyw)//' not found.')
!           end if
!         else
!             call error('NotFoundDETID', 'Keyword '//trim(keyw)//' not found.')
!         end if
!       end if
!     else
!       if (hasColumn(catTab, 'DETID')) then
!         retStr = 'DETID'
!       else if (hasColumn(catTab, 'detid')) then
!         retStr = 'detid'
!       else if (hasColumn(catTab, 'DET_ID')) then
!         retStr = 'DET_ID'
!       else if (hasColumn(catTab, 'det_id')) then
!         retStr = 'det_id'
!       else
!         if (present(noErr)) then
!           if (noErr) then
!             retStr = ''
!           else
!             call error('NotFoundDETID', 'DETID (or alike) not found.')
!           end if
!         else
!             call error('NotFoundDETID', 'DETID (or alike) not found.')
!         end if
!       end if
!     end if
! 
!   end function getDetIdStr
! 
!   !---------------------------------------------------------------------
!   function getSrcIdStr(catTab, keyw, noErr) result(retStr)
!   !---------------------------------------------------------------------
!   ! Return the column title name of SRCID.
!   ! If noErr option is specified and is true, then this does not issue error,
!   ! and returns '' if an erroneous situation happens.
! 
!     implicit none 
! 
!     character(LEN=LENOBSIDSTR) :: retStr
!     type(TableT), intent(in) :: catTab
!     character(*), intent(in), optional :: keyw
!     logical, intent(in), optional :: noErr	! Default: .false.
! 
!     if (present(keyw)) then
!       if (hasColumn(catTab, trim(keyw))) then
!         retStr = trim(keyw)
!       else
!         if (present(noErr)) then
!           if (noErr) then
!             retStr = ''
!           else
!             call error('NotFoundSRCID', 'Keyword '//trim(keyw)//' not found.')
!           end if
!         else
!             call error('NotFoundSRCID', 'Keyword '//trim(keyw)//' not found.')
!         end if
!       end if
!     else
!       if (hasColumn(catTab, 'SRCID')) then
!         retStr = 'SRCID'
!       else if (hasColumn(catTab, 'srcid')) then
!         retStr = 'srcid'
!       else if (hasColumn(catTab, 'SRC_ID')) then
!         retStr = 'SRC_ID'
!       else if (hasColumn(catTab, 'src_id')) then
!         retStr = 'src_id'
!       else
!         if (present(noErr)) then
!           if (noErr) then
!             retStr = ''
!           else
!             call error('NotFoundSRCID', 'SRCID (or alike) not found.')
!           end if
!         else
!             call error('NotFoundSRCID', 'SRCID (or alike) not found.')
!         end if
!       end if
!     end if
! 
!   end function getSrcIdStr
! 
!   !---------------------------------------------------------------------
!   function getInCatFlgStr(catTab, keyw, noErr) result(retStr)
!   !---------------------------------------------------------------------
!   ! Return the column title name of INCATFLG.
!   ! If noErr option is specified and is true (or simply not specified), then this does not issue error,
!   ! and returns '' if an erroneous situation happens.
! 
!     implicit none 
! 
!     character(LEN=LENOBSIDSTR) :: retStr
!     type(TableT), intent(in) :: catTab
!     character(*), intent(in), optional :: keyw
!     logical, intent(in), optional :: noErr	! Default: .true.
! 
!     if (present(keyw)) then
!       if (hasColumn(catTab, trim(keyw))) then
!         retStr = trim(keyw)
!       else
!         if (present(noErr)) then
!           if (noErr) then
!             retStr = ''
!           else
!             call error('NotFoundINCATFLG', 'Keyword '//trim(keyw)//' not found.')
!           end if
!         else
!             call error('NotFoundINCATFLG', 'Keyword '//trim(keyw)//' not found.')
!         end if
!       end if
!     else
!       if (hasColumn(catTab, 'INCATFLG')) then
!         retStr = 'INCATFLG'
!       else if (hasColumn(catTab, 'incatflg')) then
!         retStr = 'incatflg'
!       else
!         if (.not. present(noErr)) then
!           retStr = ''
!         else if (present(noErr)) then
!           if (noErr) then
!             retStr = ''
!           else
!             call error('NotFoundINCATFLG', 'INCATFLG (or alike) not found.')
!           end if
!         else
!             call error('NotFoundINCATFLG', 'INCATFLG (or alike) not found.')
!         end if
!       end if
!     end if
! 
!   end function getInCatFlgStr


  !---------------------------------------------------------------------
  function getColKwdStrFromKey(inTab, keyw, errName, noErr) result(retStr)
  !---------------------------------------------------------------------
  ! Return the column title name of the given KEYW.
  ! If noErr option is specified and is true, then this does not issue error,
  ! and returns '' if an erroneous situation happens.

    implicit none 

    character(LEN=LENSTRCOLNAME) :: retStr
    type(TableT), intent(in) :: inTab
    character(*), intent(in) :: keyw
    character(*), intent(in), optional :: errName
    logical, intent(in), optional :: noErr	! Default: .false.

    character(LEN=LENSTRCOLNAME) :: errNameNow

    !-------------------------------------------------------------------
    ! Start:

    if (hasColumn(inTab, trim(keyw))) then
      retStr = trim(keyw)
      ! return	! unnecessary

    else	! i.e., Error.
      if (present(noErr)) then
        if (noErr) then
          retStr = ''
          return
        end if
      end if

      if (present(errName)) then
        errNameNow = errName
      else
        errNameNow = 'NotFoundColName'//trim(keyw)
      end if
      call error(trim(errNameNow), 'Column name '//trim(keyw)//' not found.')
    end if

  end function getColKwdStrFromKey

  !---------------------------------------------------------------------
  function getColKwdStrFromAry(inTab, arKeyw, errName, noErr) result(retStr)
  !---------------------------------------------------------------------
  ! Return the column title name that matches either of the given array KEYW.
  ! If noErr option is specified and is true, then this does not issue error,
  ! and returns '' if an erroneous situation happens.

    implicit none 

    character(LEN=LENSTRCOLNAME) :: retStr
    type(TableT), intent(in) :: inTab
    character(*), intent(in) :: arKeyw(:)
    character(*), intent(in), optional :: errName
    logical, intent(in), optional :: noErr	! Default: .false.

    character(LEN=LENSTRCOLNAME) :: errNameNow
    integer :: i

    !-------------------------------------------------------------------
    ! Start:

    do i=1, size(arKeyw)
      if (trim(arKeyw(i)) == '') then
        cycle
      else if (hasColumn(inTab, trim(arKeyw(i)))) then
        retStr = trim(arKeyw(i))
        return
      else
        ! Do nothing.  Go to the next iteration.
      end if
    end do

    !! The keyword NOT found after all.
    if (present(noErr)) then
      if (noErr) then
        retStr = ''
        return
      end if
    end if

    if (present(errName)) then
      errNameNow = errName
    else
      errNameNow = 'NotFoundColName'//trim(arKeyw(1))
    end if
    call error(trim(errNameNow), 'Column name '//trim(arKeyw(1))//' (or alike) not found.')

  end function getColKwdStrFromAry


  !---------------------------------------------------------------------
  function getRaStr(catTab, keyw, isCorrPreferred) result(retStr)
  !---------------------------------------------------------------------
  ! Return the column title name of RA.

    implicit none 

    character(LEN=LENOBSIDSTR) :: retStr
    type(TableT), intent(in) :: catTab
    character(*), intent(in), optional :: keyw
    logical, intent(in), optional :: isCorrPreferred	! Default: .false.

    if (present(keyw)) then
      if (hasColumn(catTab, trim(keyw))) then
        retStr = trim(keyw)
      else
        call error('NotFoundRA', 'Keyword '//trim(keyw)//' not found.')
      end if
    else	! if (present(keyw)) then
      if (present(isCorrPreferred)) then
        if (isCorrPreferred) then
          if (hasColumn(catTab, 'RA_CORR')) then
            retStr = 'RA_CORR'
            return
          else if (hasColumn(catTab, 'ra_corr')) then
            retStr = 'ra_corr'
            return
          end if
        end if
      end if

      if (hasColumn(catTab, 'RA')) then
        retStr = 'RA'
      else if (hasColumn(catTab, 'ra')) then
        retStr = 'ra'
      else if (hasColumn(catTab, 'RA_UNC')) then
        retStr = 'RA_UNC'
      else if (hasColumn(catTab, 'ra_unc')) then
        retStr = 'ra_unc'
      else if (hasColumn(catTab, 'RA_CORR')) then
        retStr = 'RA_CORR'
      else if (hasColumn(catTab, 'ra_corr')) then
        retStr = 'ra_corr'
      else
        call error('NotFoundRA', 'RA (or alike) not found.')
      end if
    end if	! if (present(keyw)) then

  end function getRaStr

  !---------------------------------------------------------------------
  function getDecStr(catTab, keyw, isCorrPreferred) result(retStr)
  !---------------------------------------------------------------------
  ! Return the column title name of DEC.

    implicit none 

    character(LEN=LENOBSIDSTR) :: retStr
    type(TableT), intent(in) :: catTab
    character(*), intent(in), optional :: keyw
    logical, intent(in), optional :: isCorrPreferred	! Default: .false.

    if (present(keyw)) then
      if (hasColumn(catTab, trim(keyw))) then
        retStr = trim(keyw)
      else
        call error('NotFoundDEC', 'Keyword '//trim(keyw)//' not found.')
      end if
    else	! if (present(keyw)) then
      if (present(isCorrPreferred)) then
        if (isCorrPreferred) then
          if (hasColumn(catTab, 'DEC_CORR')) then
            retStr = 'DEC_CORR'
            return
          else if (hasColumn(catTab, 'dec_corr')) then
            retStr = 'dec_corr'
            return
          end if
        end if
      end if

      if (hasColumn(catTab, 'DEC')) then
        retStr = 'DEC'
      else if (hasColumn(catTab, 'dec')) then
        retStr = 'dec'
      else if (hasColumn(catTab, 'DEC_UNC')) then
        retStr = 'DEC_UNC'
      else if (hasColumn(catTab, 'dec_unc')) then
        retStr = 'dec_unc'
      else if (hasColumn(catTab, 'DEC_CORR')) then
        retStr = 'DEC_CORR'
      else if (hasColumn(catTab, 'dec_corr')) then
        retStr = 'dec_corr'
      else
        call error('NotFoundDEC', 'DEC (or alike) not found.')
      end if
    end if	! if (present(keyw)) then

  end function getDecStr


  !---------------------------------------------------------------------
  subroutine retKwdFromAnyBlockInt32FromSet(outValue, inSet, inKwd &
     , numBlock, isFound, aryValues, aryFound)
  !---------------------------------------------------------------------
  ! In some cases, one wants to get the header keyword attribute,
  ! whichever header, either primary or n-th extention, contains it.
  ! This subroutine does that job, namely it returns the value(s)
  ! of header keyword, whereever it exists in the given FITS file.
  !
  ! If the specified attribute is not found anywhere,
  ! then this subroutine raises an error in default,
  ! unless the optional parameter of either (or both) isFound or aryFound is given,
  ! in which case this instead returns .false. in isFound
  ! and/or aryFound(-1:).
  !
  ! If non-allocated pointers of aryValues(:) and aryFound(:) are given, then
  ! aryValues(-1:) returns an array of the attribute values for all the extentions,
  ! aryFound(-1:) returns a logical array of which extentions have a value.
  ! Note that the primary header comes first (index=-1),
  ! the first header, next (index=0), of which the block number in SAS is 0,
  ! and so on.
  ! In that case, outValue and numBlock (if given) are the value found
  ! first, if the keyword is found in more than one header.
  ! If you give aryValues(:) and/or aryFound(:),
  ! make sure to deallocate them after use.
  !
  ! Example:
  !   If an input FITS has the structure of up to the 3rd header with these:
  !     Primary:     OBSID = 123
  !     Header(3rd): OBSID = 456
  !   then the returned values will be
  !     outValue = 123
  !     numBlock = -1  (namely, primary)
  !     isFound  = .true.
  !     aryValues(-1:2) = (/ 123, UNDEF, UNDEF, 456 /)
  !     aryFound(-1:2)  = (/ T,   F,     F,     T   /)
  !   where UNDEF here simply means uninitialised.
  !

    implicit none 

    integer(int32), intent(out) :: outValue
    type(DataSetT), intent(in) :: inSet
    character(*),   intent(in) :: inKwd
    integer(int32), intent(inout), optional :: numBlock
    logical,        intent(out), optional :: isFound
    integer(int32), pointer, optional :: aryValues(:)	! intent(out)
    logical,        pointer, optional :: aryFound(:)	! intent(out)

    type(BlockT) :: nowBlock
    integer(int32) :: i, nBlocks, outValueNow
    logical :: isFoundAlready, doesQuitAsap
    integer(int32), allocatable :: aryValuesNow(:)
    logical,        allocatable :: aryFoundNow(:)

    !-------------------------------------------------------------------
    ! Start:

    nBlocks = numberOfBlocks(inSet)
    if (present(aryValues) .or. present(aryFound)) then
      doesQuitAsap = .false.
      if (present(aryValues)) allocate(aryValues(-1:nBlocks-1))
      if (present(aryFound))  allocate(aryFound(-1:nBlocks-1))
      allocate(aryValuesNow(-1:nBlocks-1), aryFoundNow(-1:nBlocks-1))
      aryFoundNow = .false.
    else
      doesQuitAsap = .true.
    end if

    isFoundAlready = .false.
    if (present(numBlock)) numBlock = -9999	! Default
    if (present(isFound)) isFound = .false.	! Default

    !! Primary
    if (hasAttribute(inSet, inKwd)) then
      if (dataType(attribute(inSet, inKwd)) == INTEGER_ATTRIBUTE) then
        outValue = int32Attribute(inSet, inKwd)
        if (present(numBlock)) numBlock = -1
        if (present(isFound)) isFound = .true.
        if (doesQuitAsap) then
          return
        else
          isFoundAlready   = .true.
          aryFoundNow(-1)  = .true.
          aryValuesNow(-1) = outValue
        end if
      else
        call warning('typeMismatchInAttribute' &
           , 'The attribute '//trim(inKwd) &
           //' is found in the primary header, but not Integer as supposed to be.')
      end if
    end if

    !! Following extentions
    do i = 0, nBlocks-1
      nowBlock = block(inSet, i, READ)
      if (hasAttribute(nowBlock, inKwd)) then
        if (dataType(attribute(nowBlock, inKwd)) == INTEGER_ATTRIBUTE) then
          outValueNow = int32Attribute(nowBlock, inKwd)
          if (.not. isFoundAlready) then
            outValue = outValueNow
            if (present(numBlock)) numBlock = i
            if (present(isFound)) isFound = .true.
          end if

          if (doesQuitAsap) then
            call release(nowBlock)
            return
          else
            isFoundAlready  = .true.
            aryFoundNow(i)  = .true.
            aryValuesNow(i) = outValueNow
          end if

        else
          write(errStr, '(a, i5, a)') 'The attribute '//trim(inKwd) &
             //' is found in the header' &
             , i, ', but not Integer as supposed to be.'
          call warning('typeMismatchInAttribute')
        end if
      end if
    end do

    if (present(aryValues)) aryValues = aryValuesNow
    if (present(aryFound))  aryFound  = aryFoundNow

    if (present(isFound) .or. present(aryFound)) then
      ! Do nothing.
    else
      call release(inSet)
      call error('noAttributeFound' &
         , 'The attribute '//trim(inKwd)//' is not found anywhere.')
    end if

  end subroutine retKwdFromAnyBlockInt32FromSet


  !---------------------------------------------------------------------
  subroutine retKwdFromAnyBlockInt32FromName(outValue, inSetName, inKwd &
     , numBlock, isFound, aryValues, aryFound)
  !---------------------------------------------------------------------
  ! See retKwdFromAnyBlockInt32FromSet() for detail.

    implicit none 

    integer(int32), intent(out) :: outValue
    character(*),   intent(in) :: inSetName, inKwd
    integer(int32), intent(inout), optional :: numBlock
    logical,        intent(out), optional :: isFound
    integer(int32), pointer, optional :: aryValues(:)	! intent(out)
    logical,        pointer, optional :: aryFound(:)	! intent(out)

    type(DataSetT) :: tmpSet

    !-------------------------------------------------------------------
    ! Start:

    tmpSet = dataSet(trim(inSetName), READ)
    call retKwdFromAnyBlockInt32FromSet(outValue, tmpSet, trim(inKwd) &
       , numBlock, isFound, aryValues, aryFound)
    call release(tmpSet)

  end subroutine retKwdFromAnyBlockInt32FromName


  !---------------------------------------------------------------------
  subroutine retKwdFromAnyBlockInt16FromSet(outValue, inSet, inKwd &
     , numBlock, isFound, aryValues, aryFound)
  !---------------------------------------------------------------------
  ! See retKwdFromAnyBlockInt32FromSet() for detail.

    implicit none 

    integer(int16), intent(out) :: outValue
    type(DataSetT), intent(in) :: inSet
    character(*),   intent(in) :: inKwd
    integer(int32), intent(inout), optional :: numBlock
    logical,        intent(out), optional :: isFound
    integer(int16), pointer, optional :: aryValues(:)	! intent(out)
    logical,        pointer, optional :: aryFound(:)	! intent(out)

    type(BlockT) :: nowBlock
    integer(int32) :: i, nBlocks
    integer(int16) :: outValueNow
    logical :: isFoundAlready, doesQuitAsap
    integer(int16), allocatable :: aryValuesNow(:)
    logical,        allocatable :: aryFoundNow(:)

    !-------------------------------------------------------------------
    ! Start:

    nBlocks = numberOfBlocks(inSet)
    if (present(aryValues) .or. present(aryFound)) then
      doesQuitAsap = .false.
      if (present(aryValues)) allocate(aryValues(-1:nBlocks-1))
      if (present(aryFound))  allocate(aryFound(-1:nBlocks-1))
      allocate(aryValuesNow(-1:nBlocks-1), aryFoundNow(-1:nBlocks-1))
      aryFoundNow = .false.
    else
      doesQuitAsap = .true.
    end if

    isFoundAlready = .false.
    if (present(numBlock)) numBlock = -9999	! Default
    if (present(isFound)) isFound = .false.	! Default

    !! Primary
    if (hasAttribute(inSet, inKwd)) then
      if (dataType(attribute(inSet, inKwd)) == INTEGER_ATTRIBUTE) then
        outValue = int16Attribute(inSet, inKwd)
        if (present(numBlock)) numBlock = -1
        if (present(isFound)) isFound = .true.
        if (doesQuitAsap) then
          return
        else
          isFoundAlready   = .true.
          aryFoundNow(-1)  = .true.
          aryValuesNow(-1) = outValue
        end if
      else
        call warning('typeMismatchInAttribute' &
           , 'The attribute '//trim(inKwd) &
           //' is found in the primary header, but not Integer as supposed to be.')
      end if
    end if

    !! Following extentions
    do i = 0, nBlocks-1
      nowBlock = block(inSet, i, READ)
      if (hasAttribute(nowBlock, inKwd)) then
        if (dataType(attribute(nowBlock, inKwd)) == INTEGER_ATTRIBUTE) then
          outValueNow = int16Attribute(nowBlock, inKwd)
          if (.not. isFoundAlready) then
            outValue = outValueNow
            if (present(numBlock)) numBlock = i
            if (present(isFound)) isFound = .true.
          end if

          if (doesQuitAsap) then
            call release(nowBlock)
            return
          else
            isFoundAlready  = .true.
            aryFoundNow(i)  = .true.
            aryValuesNow(i) = outValueNow
          end if

        else
          write(errStr, '(a, i5, a)') 'The attribute '//trim(inKwd) &
             //' is found in the header' &
             , i, ', but not Integer as supposed to be.'
          call warning('typeMismatchInAttribute')
        end if
      end if
    end do

    if (present(aryValues)) aryValues = aryValuesNow
    if (present(aryFound))  aryFound  = aryFoundNow

    if (present(isFound) .or. present(aryFound)) then
      ! Do nothing.
    else
      call release(inSet)
      call error('noAttributeFound' &
         , 'The attribute '//trim(inKwd)//' is not found anywhere.')
    end if

  end subroutine retKwdFromAnyBlockInt16FromSet


  !---------------------------------------------------------------------
  subroutine retKwdFromAnyBlockInt16FromName(outValue, inSetName, inKwd &
     , numBlock, isFound, aryValues, aryFound)
  !---------------------------------------------------------------------
  ! See retKwdFromAnyBlockInt32FromSet() for detail.

    implicit none 

    integer(int16), intent(out) :: outValue
    character(*),   intent(in) :: inSetName, inKwd
    integer(int32), intent(inout), optional :: numBlock
    logical,        intent(out), optional :: isFound
    integer(int16), pointer, optional :: aryValues(:)	! intent(out)
    logical,        pointer, optional :: aryFound(:)	! intent(out)

    type(DataSetT) :: tmpSet

    !-------------------------------------------------------------------
    ! Start:

    tmpSet = dataSet(trim(inSetName), READ)
    call retKwdFromAnyBlockInt16FromSet(outValue, tmpSet, trim(inKwd) &
       , numBlock, isFound, aryValues, aryFound)
    call release(tmpSet)

  end subroutine retKwdFromAnyBlockInt16FromName


  !---------------------------------------------------------------------
  subroutine retKwdFromAnyBlockInt8FromSet(outValue, inSet, inKwd &
     , numBlock, isFound, aryValues, aryFound)
  !---------------------------------------------------------------------
  ! See retKwdFromAnyBlockInt32FromSet() for detail.

    implicit none 

    integer(int8), intent(out) :: outValue
    type(DataSetT), intent(in) :: inSet
    character(*),   intent(in) :: inKwd
    integer(int32), intent(inout), optional :: numBlock
    logical,        intent(out), optional :: isFound
    integer(int8), pointer, optional :: aryValues(:)	! intent(out)
    logical,        pointer, optional :: aryFound(:)	! intent(out)

    type(BlockT) :: nowBlock
    integer(int32) :: i, nBlocks
    integer(int8) :: outValueNow
    logical :: isFoundAlready, doesQuitAsap
    integer(int8), allocatable :: aryValuesNow(:)
    logical,        allocatable :: aryFoundNow(:)

    !-------------------------------------------------------------------
    ! Start:

    nBlocks = numberOfBlocks(inSet)
    if (present(aryValues) .or. present(aryFound)) then
      doesQuitAsap = .false.
      if (present(aryValues)) allocate(aryValues(-1:nBlocks-1))
      if (present(aryFound))  allocate(aryFound(-1:nBlocks-1))
      allocate(aryValuesNow(-1:nBlocks-1), aryFoundNow(-1:nBlocks-1))
      aryFoundNow = .false.
    else
      doesQuitAsap = .true.
    end if

    isFoundAlready = .false.
    if (present(numBlock)) numBlock = -9999	! Default
    if (present(isFound)) isFound = .false.	! Default

    !! Primary
    if (hasAttribute(inSet, inKwd)) then
      if (dataType(attribute(inSet, inKwd)) == INTEGER_ATTRIBUTE) then
        outValue = int8Attribute(inSet, inKwd)
        if (present(numBlock)) numBlock = -1
        if (present(isFound)) isFound = .true.
        if (doesQuitAsap) then
          return
        else
          isFoundAlready   = .true.
          aryFoundNow(-1)  = .true.
          aryValuesNow(-1) = outValue
        end if
      else
        call warning('typeMismatchInAttribute' &
           , 'The attribute '//trim(inKwd) &
           //' is found in the primary header, but not Integer as supposed to be.')
      end if
    end if

    !! Following extentions
    do i = 0, nBlocks-1
      nowBlock = block(inSet, i, READ)
      if (hasAttribute(nowBlock, inKwd)) then
        if (dataType(attribute(nowBlock, inKwd)) == INTEGER_ATTRIBUTE) then
          outValueNow = int8Attribute(nowBlock, inKwd)
          if (.not. isFoundAlready) then
            outValue = outValueNow
            if (present(numBlock)) numBlock = i
            if (present(isFound)) isFound = .true.
          end if

          if (doesQuitAsap) then
            call release(nowBlock)
            return
          else
            isFoundAlready  = .true.
            aryFoundNow(i)  = .true.
            aryValuesNow(i) = outValueNow
          end if

        else
          write(errStr, '(a, i5, a)') 'The attribute '//trim(inKwd) &
             //' is found in the header' &
             , i, ', but not Integer as supposed to be.'
          call warning('typeMismatchInAttribute')
        end if
      end if
    end do

    if (present(aryValues)) aryValues = aryValuesNow
    if (present(aryFound))  aryFound  = aryFoundNow

    if (present(isFound) .or. present(aryFound)) then
      ! Do nothing.
    else
      call release(inSet)
      call error('noAttributeFound' &
         , 'The attribute '//trim(inKwd)//' is not found anywhere.')
    end if

  end subroutine retKwdFromAnyBlockInt8FromSet


  !---------------------------------------------------------------------
  subroutine retKwdFromAnyBlockInt8FromName(outValue, inSetName, inKwd &
     , numBlock, isFound, aryValues, aryFound)
  !---------------------------------------------------------------------
  ! See retKwdFromAnyBlockInt32FromSet() for detail.

    implicit none 

    integer(int8), intent(out) :: outValue
    character(*),   intent(in) :: inSetName, inKwd
    integer(int32), intent(inout), optional :: numBlock
    logical,        intent(out), optional :: isFound
    integer(int8), pointer, optional :: aryValues(:)	! intent(out)
    logical,        pointer, optional :: aryFound(:)	! intent(out)

    type(DataSetT) :: tmpSet

    !-------------------------------------------------------------------
    ! Start:

    tmpSet = dataSet(trim(inSetName), READ)
    call retKwdFromAnyBlockInt8FromSet(outValue, tmpSet, trim(inKwd) &
       , numBlock, isFound, aryValues, aryFound)
    call release(tmpSet)

  end subroutine retKwdFromAnyBlockInt8FromName


  !---------------------------------------------------------------------
  subroutine retKwdFromAnyBlockDblFromSet(outValue, inSet, inKwd &
     , numBlock, isFound, aryValues, aryFound)
  !---------------------------------------------------------------------
  ! See retKwdFromAnyBlockInt32FromSet() for detail.

    implicit none 

    real(double),   intent(out) :: outValue
    type(DataSetT), intent(in) :: inSet
    character(*),   intent(in) :: inKwd
    integer(int32), intent(inout), optional :: numBlock
    logical,        intent(out), optional :: isFound
    real(double),   pointer, optional :: aryValues(:)	! intent(out)
    logical,        pointer, optional :: aryFound(:)	! intent(out)

    type(BlockT) :: nowBlock
    integer(int32) :: i, nBlocks
    real(double)   :: outValueNow
    logical :: isFoundAlready, doesQuitAsap
    real(double),   allocatable :: aryValuesNow(:)
    logical,        allocatable :: aryFoundNow(:)
    integer :: datType

    !-------------------------------------------------------------------
    ! Start:

    nBlocks = numberOfBlocks(inSet)
    if (present(aryValues) .or. present(aryFound)) then
      doesQuitAsap = .false.
      if (present(aryValues)) allocate(aryValues(-1:nBlocks-1))
      if (present(aryFound))  allocate(aryFound(-1:nBlocks-1))
      allocate(aryValuesNow(-1:nBlocks-1), aryFoundNow(-1:nBlocks-1))
      aryFoundNow = .false.
    else
      doesQuitAsap = .true.
    end if

    isFoundAlready = .false.
    if (present(numBlock)) numBlock = -9999	! Default
    if (present(isFound)) isFound = .false.	! Default

    !! Primary
    if (hasAttribute(inSet, inKwd)) then
      datType = dataType(attribute(inSet, inKwd))
      if ((datType == REAL_ATTRIBUTE) .or. (datType == INTEGER_ATTRIBUTE)) then
        if (datType == INTEGER_ATTRIBUTE) then
          call warning('typeMismatchInAttribute' &
             , 'The attribute '//trim(inKwd) &
             //' is found in the primary header, &
               &however the data type is Integer, &
               &not Real as supposed to be.  Continued.')
        end if

        outValue = real64Attribute(inSet, inKwd)
        if (present(numBlock)) numBlock = -1
        if (present(isFound)) isFound = .true.
        if (doesQuitAsap) then
          return
        else
          isFoundAlready   = .true.
          aryFoundNow(-1)  = .true.
          aryValuesNow(-1) = outValue
        end if
      else
        call warning('typeMismatchInAttribute' &
           , 'The attribute '//trim(inKwd) &
           //' is found in the primary header,&
             & but not Real (nor Integer) as supposed to be.  Skip.')
      end if
    end if

    !! Following extentions
    do i = 0, nBlocks-1
      nowBlock = block(inSet, i, READ)
      if (hasAttribute(nowBlock, inKwd)) then
        datType = dataType(attribute(nowBlock, inKwd))
        if ((datType == REAL_ATTRIBUTE) .or. (datType == INTEGER_ATTRIBUTE)) then
          if (datType == INTEGER_ATTRIBUTE) then
            write(errStr, '(a, i5, a)') 'The attribute '//trim(inKwd) &
               //' is found in the header' &
               , i, ', however the data type is Integer, &
                 &not Real as supposed to be.  Continued.'
            call warning('typeMismatchInAttribute')
          end if

          outValueNow = real64Attribute(nowBlock, inKwd)
          if (.not. isFoundAlready) then
            outValue = outValueNow
            if (present(numBlock)) numBlock = i
            if (present(isFound)) isFound = .true.
          end if

          if (doesQuitAsap) then
            call release(nowBlock)
            return
          else
            isFoundAlready  = .true.
            aryFoundNow(i)  = .true.
            aryValuesNow(i) = outValueNow
          end if

        else
          write(errStr, '(a, i5, a)') 'The attribute '//trim(inKwd) &
             //' is found in the header' &
             , i, ', but not Real (nor Integer) as supposed to be.  Skip.'
          call warning('typeMismatchInAttribute')
        end if
      end if
    end do

    if (present(aryValues)) aryValues = aryValuesNow
    if (present(aryFound))  aryFound  = aryFoundNow

    if (present(isFound) .or. present(aryFound)) then
      ! Do nothing.
    else
      call release(inSet)
      call error('noAttributeFound' &
         , 'The attribute '//trim(inKwd)//' is not found anywhere.')
    end if

  end subroutine retKwdFromAnyBlockDblFromSet


  !---------------------------------------------------------------------
  subroutine retKwdFromAnyBlockDblFromName(outValue, inSetName, inKwd &
     , numBlock, isFound, aryValues, aryFound)
  !---------------------------------------------------------------------
  ! See retKwdFromAnyBlockInt32FromSet() for detail.

    implicit none 

    real(double),   intent(out) :: outValue
    character(*),   intent(in) :: inSetName, inKwd
    integer(int32), intent(inout), optional :: numBlock
    logical,        intent(out), optional :: isFound
    real(double),   pointer, optional :: aryValues(:)	! intent(out)
    logical,        pointer, optional :: aryFound(:)	! intent(out)

    type(DataSetT) :: tmpSet

    !-------------------------------------------------------------------
    ! Start:

    tmpSet = dataSet(trim(inSetName), READ)
    call retKwdFromAnyBlockDblFromSet(outValue, tmpSet, trim(inKwd) &
       , numBlock, isFound, aryValues, aryFound)
    call release(tmpSet)

  end subroutine retKwdFromAnyBlockDblFromName


  !---------------------------------------------------------------------
  subroutine retKwdFromAnyBlockSglFromSet(outValue, inSet, inKwd &
     , numBlock, isFound, aryValues, aryFound)
  !---------------------------------------------------------------------
  ! See retKwdFromAnyBlockInt32FromSet() for detail.

    implicit none 

    real(single),   intent(out) :: outValue
    type(DataSetT), intent(in) :: inSet
    character(*),   intent(in) :: inKwd
    integer(int32), intent(inout), optional :: numBlock
    logical,        intent(out), optional :: isFound
    real(single),   pointer, optional :: aryValues(:)	! intent(out)
    logical,        pointer, optional :: aryFound(:)	! intent(out)

    type(BlockT) :: nowBlock
    integer(int32) :: i, nBlocks
    real(single)   :: outValueNow
    logical :: isFoundAlready, doesQuitAsap
    real(single),   allocatable :: aryValuesNow(:)
    logical,        allocatable :: aryFoundNow(:)
    integer :: datType

    !-------------------------------------------------------------------
    ! Start:

    nBlocks = numberOfBlocks(inSet)
    if (present(aryValues) .or. present(aryFound)) then
      doesQuitAsap = .false.
      if (present(aryValues)) allocate(aryValues(-1:nBlocks-1))
      if (present(aryFound))  allocate(aryFound(-1:nBlocks-1))
      allocate(aryValuesNow(-1:nBlocks-1), aryFoundNow(-1:nBlocks-1))
      aryFoundNow = .false.
    else
      doesQuitAsap = .true.
    end if

    isFoundAlready = .false.
    if (present(numBlock)) numBlock = -9999	! Default
    if (present(isFound)) isFound = .false.	! Default

    !! Primary
    if (hasAttribute(inSet, inKwd)) then
      datType = dataType(attribute(inSet, inKwd))
      if ((datType == REAL_ATTRIBUTE) .or. (datType == INTEGER_ATTRIBUTE)) then
        if (datType == INTEGER_ATTRIBUTE) then
          call warning('typeMismatchInAttribute' &
             , 'The attribute '//trim(inKwd) &
             //' is found in the primary header, &
               &however the data type is Integer, &
               &not Real as supposed to be.  Continued.')
        end if

        outValue = real32Attribute(inSet, inKwd)
        if (present(numBlock)) numBlock = -1
        if (present(isFound)) isFound = .true.
        if (doesQuitAsap) then
          return
        else
          isFoundAlready   = .true.
          aryFoundNow(-1)  = .true.
          aryValuesNow(-1) = outValue
        end if
      else
        call warning('typeMismatchInAttribute' &
           , 'The attribute '//trim(inKwd) &
           //' is found in the primary header,&
             & but not Real (nor Integer) as supposed to be.  Skip.')
      end if
    end if

    !! Following extentions
    do i = 0, nBlocks-1
      nowBlock = block(inSet, i, READ)
      if (hasAttribute(nowBlock, inKwd)) then
        datType = dataType(attribute(nowBlock, inKwd))
        if ((datType == REAL_ATTRIBUTE) .or. (datType == INTEGER_ATTRIBUTE)) then
          if (datType == INTEGER_ATTRIBUTE) then
            write(errStr, '(a, i5, a)') 'The attribute '//trim(inKwd) &
               //' is found in the header' &
               , i, ', however the data type is Integer, &
                 &not Real as supposed to be.  Continued.'
            call warning('typeMismatchInAttribute')
          end if

          outValueNow = real32Attribute(nowBlock, inKwd)
          if (.not. isFoundAlready) then
            outValue = outValueNow
            if (present(numBlock)) numBlock = i
            if (present(isFound)) isFound = .true.
          end if

          if (doesQuitAsap) then
            call release(nowBlock)
            return
          else
            isFoundAlready  = .true.
            aryFoundNow(i)  = .true.
            aryValuesNow(i) = outValueNow
          end if

        else
          write(errStr, '(a, i5, a)') 'The attribute '//trim(inKwd) &
             //' is found in the header' &
             , i, ', but not Real (nor Integer) as supposed to be.  Skip.'
          call warning('typeMismatchInAttribute')
        end if
      end if
    end do

    if (present(aryValues)) aryValues = aryValuesNow
    if (present(aryFound))  aryFound  = aryFoundNow

    if (present(isFound) .or. present(aryFound)) then
      ! Do nothing.
    else
      call release(inSet)
      call error('noAttributeFound' &
         , 'The attribute '//trim(inKwd)//' is not found anywhere.')
    end if

  end subroutine retKwdFromAnyBlockSglFromSet


  !---------------------------------------------------------------------
  subroutine retKwdFromAnyBlockSglFromName(outValue, inSetName, inKwd &
     , numBlock, isFound, aryValues, aryFound)
  !---------------------------------------------------------------------
  ! See retKwdFromAnyBlockInt32FromSet() for detail.

    implicit none 

    real(single),   intent(out) :: outValue
    character(*),   intent(in) :: inSetName, inKwd
    integer(int32), intent(inout), optional :: numBlock
    logical,        intent(out), optional :: isFound
    real(single),   pointer, optional :: aryValues(:)	! intent(out)
    logical,        pointer, optional :: aryFound(:)	! intent(out)

    type(DataSetT) :: tmpSet

    !-------------------------------------------------------------------
    ! Start:

    tmpSet = dataSet(trim(inSetName), READ)
    call retKwdFromAnyBlockSglFromSet(outValue, tmpSet, trim(inKwd) &
       , numBlock, isFound, aryValues, aryFound)
    call release(tmpSet)

  end subroutine retKwdFromAnyBlockSglFromName


  !---------------------------------------------------------------------
  subroutine retKwdFromAnyBlockCharFromSet(outValue, inSet, inKwd &
     , numBlock, isFound, aryValues, aryFound)
  !---------------------------------------------------------------------
  ! See retKwdFromAnyBlockInt32FromSet() for detail.

    implicit none 

    character(*),   intent(out) :: outValue
    type(DataSetT), intent(in) :: inSet
    character(*),   intent(in) :: inKwd
    integer(int32), intent(inout), optional :: numBlock
    logical,        intent(out), optional :: isFound
    character(*),   pointer, optional :: aryValues(:)	! intent(out)
    logical,        pointer, optional :: aryFound(:)	! intent(out)

    type(BlockT) :: nowBlock
    integer(int32) :: i, nBlocks
    character(len(outValue))   :: outValueNow
    logical :: isFoundAlready, doesQuitAsap
    !!character(len(aryValues)),   allocatable :: aryValuesNow(:)
    character(len=255),   allocatable :: aryValuesNow(:)
    logical,        allocatable :: aryFoundNow(:)

    !-------------------------------------------------------------------
    ! Start:

    nBlocks = numberOfBlocks(inSet)
    if (present(aryValues) .or. present(aryFound)) then
      doesQuitAsap = .false.
      if (present(aryValues)) allocate(aryValues(-1:nBlocks-1))
      if (present(aryFound))  allocate(aryFound(-1:nBlocks-1))
      allocate(aryValuesNow(-1:nBlocks-1), aryFoundNow(-1:nBlocks-1))
      aryFoundNow = .false.
    else
      doesQuitAsap = .true.
    end if

    isFoundAlready = .false.
    if (present(numBlock)) numBlock = -9999	! Default
    if (present(isFound)) isFound = .false.	! Default

    !! Primary
    if (hasAttribute(inSet, inKwd)) then
      if (dataType(attribute(inSet, inKwd)) == STRING_ATTRIBUTE) then
        outValue = stringAttribute(inSet, inKwd)
        if (present(numBlock)) numBlock = -1
        if (present(isFound)) isFound = .true.
        if (doesQuitAsap) then
          return
        else
          isFoundAlready   = .true.
          aryFoundNow(-1)  = .true.
          aryValuesNow(-1) = outValue
        end if
      else
        call warning('typeMismatchInAttribute' &
           , 'The attribute '//trim(inKwd) &
           //' is found in the primary header, but not Integer as supposed to be.')
      end if
    end if

    !! Following extentions
    do i = 0, nBlocks-1
      nowBlock = block(inSet, i, READ)
      if (hasAttribute(nowBlock, inKwd)) then
        if (dataType(attribute(nowBlock, inKwd)) == STRING_ATTRIBUTE) then
          outValueNow = stringAttribute(nowBlock, inKwd)
          if (.not. isFoundAlready) then
            outValue = outValueNow
            if (present(numBlock)) numBlock = i
            if (present(isFound)) isFound = .true.
          end if

          if (doesQuitAsap) then
            call release(nowBlock)
            return
          else
            isFoundAlready  = .true.
            aryFoundNow(i)  = .true.
            aryValuesNow(i) = outValueNow
          end if

        else
          write(errStr, '(a, i5, a)') 'The attribute '//trim(inKwd) &
             //' is found in the header' &
             , i, ', but not Integer as supposed to be.'
          call warning('typeMismatchInAttribute')
        end if
      end if
    end do

    if (present(aryValues)) aryValues = aryValuesNow
    if (present(aryFound))  aryFound  = aryFoundNow

    if (present(isFound) .or. present(aryFound)) then
      ! Do nothing.
    else
      call release(inSet)
      call error('noAttributeFound' &
         , 'The attribute '//trim(inKwd)//' is not found anywhere.')
    end if

  end subroutine retKwdFromAnyBlockCharFromSet


  !---------------------------------------------------------------------
  subroutine retKwdFromAnyBlockCharFromName(outValue, inSetName, inKwd &
     , numBlock, isFound, aryValues, aryFound)
  !---------------------------------------------------------------------
  ! See retKwdFromAnyBlockInt32FromSet() for detail.

    implicit none 

    character(*),   intent(out) :: outValue
    character(*),   intent(in) :: inSetName, inKwd
    integer(int32), intent(inout), optional :: numBlock
    logical,        intent(out), optional :: isFound
    character(*),   pointer, optional :: aryValues(:)	! intent(out)
    logical,        pointer, optional :: aryFound(:)	! intent(out)

    type(DataSetT) :: tmpSet

    !-------------------------------------------------------------------
    ! Start:

    tmpSet = dataSet(trim(inSetName), READ)
    call retKwdFromAnyBlockCharFromSet(outValue, tmpSet, trim(inKwd) &
       , numBlock, isFound, aryValues, aryFound)
    call release(tmpSet)

  end subroutine retKwdFromAnyBlockCharFromName


end module ingestsrcnames_aux

