! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: ingestsrcnames_mod
!
! Authors:  Masaaki Sakano - SSC/LUX
!           Ian Stewart - SSC/LUX (originally, extractnames_mod.f90)
!
! Description: This task extracts catalog source names (IAUNAME), DETID, OBS_ID for a given observation from the source catalogue (such as, 2XMM) and ingests them, as well as INCATFLG, into the observation-summary source list.
!
! Comments: 
!
! 2006/06 : XMMSRCNAME -> IAUNAME
!           SUM_SRC_NUM -> SRC_NUM
!           OBSID is OK, now
!           Copy DET_ID in the each-obs source file from the Catalogue.
! 2007/05 : The input general catalogue file must have a SRC table 'SRCLIST'
!	that has the columns IAUNAME(XMMSRCNAME), DET_ID(DETID),
!                            SRC_NUM(SUM_SRC_NUM), OBSID(OBS_ID)
! 2008/06 : Major changes:
!	* The filename from extractnames_mod.f90 into ingestsrcnames_mod.f90
!	* Librarisation of most of functions (to ingestsrcnames_aux.f90).
!	* Made obsid command-line arg unnecessary.
!
! TODO:
!
!-
module ingestsrcnames_mod

  use dal
  use param  
  use errorHandling
!  use datafilter

  use dal_aux			! for readColDataToFixed()
  use ssc_misc
  use array_utils

  use ingestsrcnames_aux

  implicit none 

  private 

  ! Command-line parameters:
  character(255) ::&
    ArgCatSetTabName,&
    ArgSrcListSetTabName,&
    ArgNewSrcListSetName
  logical(bool) ::&
    ArgWithNewSrcListSet,&
    ArgWithObsId
  integer(int32) ::&
    ArgObsId

  public ingestsrcnames

contains 

  !---------------------------------------------------------------------
  subroutine ingestsrcnames
  !---------------------------------------------------------------------

    implicit none 

    integer, parameter :: lengthOfString = 22	! "2XMMp J890123.5-780012"
    ! integer :: arDimCatIAUName(0:0), arDimSumIAUName(0:0)
    ! character(lengthOfString) :: blankString, nameStr
    type(DataSetT) :: catSet, sumSet
    ! type(TableT)   :: catTab, sumTab
    ! integer(int32), pointer :: catObsIdData(:), &
    !      catDetIdData(:), sumDetIdData(:), catSrcNumData(:), sumSrcNumData(:)
    ! integer(int32), pointer, dimension(:) :: catObsIdDataInt
    ! character(len=10), pointer :: catObsIdDataStr(:)
    ! type(ColumnT) :: catSrcNameCol, sumCatSrcNameCol, obsidCol

    character(LENSTRFILENAME) :: srcListSetName, catSetName
    character(LENONEFITSLINE) :: srcListTabName, catTabName
    ! integer(int32) :: sumRow, catRow, numMatches
    ! integer :: dattype, i, nrow
    ! character(LEN=10) :: cellObsIdStr
    ! character(LEN=10) :: srcNumStr
    ! character(LEN=128) :: colNameObsIdStr, colNameSrcNumStr, &
    !      colNameDetIdStr, colNameIAUNameStr
    ! logical(bool), pointer :: sumInCatData(:)
    logical :: noColonFoundCat, noColonFoundSrcList

    !-------------------------------------------------------------------
    ! Start:

    call readParameters

    call splitSetTabName(ArgCatSetTabName, catSetName, catTabName &
       , noColonFoundCat)	! in ssclib/dal_aux
    call splitSetTabName(ArgSrcListSetTabName, srcListSetName, srcListTabName &
       , noColonFoundSrcList)	! in ssclib/dal_aux

    ! if (ArgWithNewSrcListSet) then
    !   call copyFile(srcListSetName, ArgNewSrcListSetName, isClobber=clobber())	! in ingestsrcnames_aux
    !   !call copyFits(srcListSetName, ArgNewSrcListSetName, isClobber=clobber())	! in ingestsrcnames_aux
    ! else
    !   ArgNewSrcListSetName = trim(srcListSetName)
    ! end if

    ! blankString = repeat(' ', lengthOfString)

    !! Read from each OBS source catalogue
    if (ArgWithNewSrcListSet) then
      sumSet = clone(srcListSetName, ArgNewSrcListSetName, MODIFY)
      !sumSet = clone(srcListSetName, ArgNewSrcListSetName, CREATE)	! Invalid
    else
      !sumSet = dataSet(newSumSetName, MODIFY)
      sumSet = dataSet(srcListSetName, MODIFY)
    end if
    ! sumTab = table(sumSet, 'SRCLIST')

    !! Read from all the XMM source catalogue (such as 2XMM)
    catSet = dataSet(catSetName, READ)
    ! call getCat4ColsFrom(catSet, catObsIdDataInt &
    !  , catSrcNumData, catDetIdData, catSrcNameCol)	! catObsIdDataInt is allocated.

    if (verbosity(AppMsg) >= Noisy) then
      call message(AppMsg, Noisy, 'Cataglogue name is '//trim(catSetName))
      if (.not. noColonFoundCat) then
        call message(AppMsg, Noisy, 'Cataglogue Table name is '//trim(catTabName))
      else
        call message(AppMsg, Noisy, 'No cataglogue Table name is given.')
      end if
  
      call message(AppMsg, Noisy, 'SRC-List name is '//trim(srcListSetName))
      if (.not. noColonFoundSrcList) then
        call message(AppMsg, Noisy, 'SRC-List Table name is '//trim(srcListTabName))
      else
        call message(AppMsg, Noisy, 'No SRC-List Table name is given (Def: '&
           //trim(DEFAULTSRCTBLNAME)//').')
      end if
      if (ArgWithNewSrcListSet) then
        call message(AppMsg, Noisy, 'Output SRC-List name is '//trim(ArgNewSrcListSetName))
      else
        call message(AppMsg, Noisy, 'Input SRC-List is overwritten.')
      end if
    end if

    if (ArgWithObsId) then
      if (noColonFoundCat .and. noColonFoundSrcList) then
        call addNew3Cols(sumSet, catSet, obsId=ArgObsId)
      else if ((.not. noColonFoundCat) .and.        noColonFoundSrcList)  then
        call addNew3Cols(sumSet, catSet, obsId=ArgObsId, catTabName=catTabName)
      else if (       noColonFoundCat  .and. (.not. noColonFoundSrcList)) then
        call addNew3Cols(sumSet, catSet, obsId=ArgObsId, sumTabName=srcListTabName)
      else	! .not. (noColonFoundCat .and. noColonFoundSrcList)
        call addNew3Cols(sumSet, catSet, obsId=ArgObsId &
           , sumTabName=srcListTabName, catTabName=catTabName)
      end if
    else
      if (noColonFoundCat .and. noColonFoundSrcList) then
print *, 'DEBUG: start ... 001000a ...'
        call addNew3Cols(sumSet, catSet)
      else if ((.not. noColonFoundCat) .and.        noColonFoundSrcList)  then
        call addNew3Cols(sumSet, catSet, catTabName=catTabName)
      else if (       noColonFoundCat  .and. (.not. noColonFoundSrcList)) then
        call addNew3Cols(sumSet, catSet, sumTabName=srcListTabName)
      else	! .not. (noColonFoundCat .and. noColonFoundSrcList)
        call addNew3Cols(sumSet, catSet &
           , sumTabName=srcListTabName, catTabName=catTabName)
      end if
    end if

    !! if (blockType(catSet, 0) == TABLE_BLOCK) then
    !!   catTab = table(catSet, 0)
    !! else if (blockType(catSet, 1) == TABLE_BLOCK) then
    !!   catTab = table(catSet, 1)
    !! else
    !!   call warning('NoSrcTable', 'Source table not found in 1st Extension or before.  Try SRCLIST.')
    !!   catTab = table(catSet, 'SRCLIST')
    !!   !! Or, possibly can read each extension till reaching the 1st Table?
    !! end if
    !! 
    !! colNameObsIdStr = getObsIdStr(catTab)
    !! 
    !! obsidCol = column(catTab, trim(colNameObsIdStr), READ)
    !! dattype = dataType(obsidCol)
    !! select case (dattype)
    !! case (INTEGER32, INTEGER16)
    !!   catObsIdData => int32Data(obsidCol)
    !! 
    !! case (STRING)
    !!   nrow = numberOfRows(catTab)-1
    !!   allocate(catObsIdDataStr(0:nrow))
    !!   allocate(catObsIdDataInt(0:nrow))
    !! 
    !!   call readColDataToFixed(catTab, trim(colNameObsIdStr),  catObsIdDataStr)
    !!   do i=0,nrow
    !!     read(catObsIdDataStr(i), *) catObsIdDataInt(i)
    !!   end do
    !!   !!   readColDataToFixed() is defined in dal_aux in ssc_lib,
    !!   !! which is an interface to dal with the use of stringCell() (as below).
    !!   !! Therefore it is very slow...  No proper (and fast) DAL function
    !!   !! exists for this purpose (to read string columns in one go).
    !!   ! do i=0,nrow
    !!   !   cellObsIdStr = stringCell(obsidCol, i)
    !!   !   read(cellObsIdStr, *) catObsIdDataInt(i)
    !!   ! end do
    !!   catObsIdData => catObsIdDataInt
    !! 
    !! case default
    !!   call error('WrongDataTypeObsId', 'OBS_ID type is wrong.')
    !! end select
    !! 
    !! colNameSrcNumStr = getSrcNumStr(catTab)
    !! catSrcNumData => int32Data(column(catTab, colNameSrcNumStr))
    !! 
    !! colNameDetIdStr  = getDetIdStr(catTab)
    !! catDetIdData     => int32Data(column(catTab, colNameDetIdStr))
    !! 
    !! colNameIAUNameStr = getIAUNameStr(catTab)
    !! catSrcNameCol = column(catTab, colNameIAUNameStr)
    !! arDimCatIAUName = dimensions(catSrcNameCol)
    !
    !call addNew3ColsWithCat(sumTab &
    ! , catObsIdDataInt &
    ! , catSrcNumData, catDetIdData, catSrcNameCol &
    ! , ArgObsId)
    !
    !! sumSrcNumData => int32Data(column(sumTab, 'SRC_NUM'))
    !!
    !! !! Addition to the source catalogue of each OBS_ID
    !! if (hasColumn(sumTab, 'IAUNAME')) then	! Force 'IAUNAME'
    !!   sumCatSrcNameCol = column(   sumTab, 'IAUNAME')
    !!   arDimSumIAUName = dimensions(sumCatSrcNameCol)
    !!   if (arDimSumIAUName(0) < arDimCatIAUName(0)) then
    !!     call warning('smallerDimensionForSumIauname',&
    !!          'IAUNAME in output Summary may be trimmed (due to shortage of the cell size).')
    !!   end if
    !! 
    !! else
    !!   arDimSumIAUName(0) = lengthOfString
    !!   sumCatSrcNameCol = addColumn(sumTab, 'IAUNAME', STRING, &
    !!        comment='Source name', dimensions=arDimSumIAUName)
    !! end if
    !! 
    !! if (hasColumn(sumTab, 'DETID')) then	! Force 'DETID'
    !!   sumDetIdData => int32Data(column(   sumTab, 'DETID'))
    !! else
    !!   sumDetIdData => int32Data(addColumn(sumTab, 'DETID', INTEGER32))
    !! end if
    !! 
    !! if (hasColumn(sumTab, 'INCATFLG')) then
    !!   sumInCatData => boolData(column(   sumTab, 'INCATFLG'))
    !! else
    !!   sumInCatData => boolData(addColumn(sumTab, 'INCATFLG', BOOLEAN))
    !! end if
    !! 
    !! 
    !! sumInCatData = .true. ! default
    !! do sumRow = 0, numberOfRows(sumTab) - 1
    !!   numMatches = count(catObsIdData==ArgObsId&
    !!     .and. catSrcNumData==sumSrcNumData(sumRow))
    !! 
    !!   if (numMatches > 1) then
    !!     srcNumStr = intToStr(sumSrcNumData(sumRow))
    !! 
    !!     write(errstr, '(a, i10.10)') 'More than one entry was found &
    !!       &in the catalog for source number '//trim(srcNumStr)//' in &
    !!       &the summary source list of OBS-ID ', ArgObsId
!! call error('tooManyMatches')
    !!   end if
    !! 
    !!   if (numMatches <= 0) then
    !!     srcNumStr = intToStr(sumSrcNumData(sumRow))
    !! 
    !!     write(errstr, '(a, i10.10)') 'No entry was found in the catalog &
    !!       &for source number '//trim(srcNumStr)//' in the summary source &
    !!       &list of OBS-ID ', ArgObsId
    !!     call warning('noMatchingSrc')
    !!     call setStringCell(sumCatSrcNameCol, sumRow, blankString)
    !! 
    !!     sumInCatData(sumRow) = .false.
    !! 
    !!   else
    !!     catRow = smallestTrueIndex(catObsIdData==ArgObsId&
    !!       .and. catSrcNumData==sumSrcNumData(sumRow)) - 1
    !! 
    !!     sumDetIdData(sumRow) = catDetIdData(catRow)	! Added DETID
    !! 
    !!     nameStr = '' ! prevents funny characters at the end if the string
    !!     ! in catSrcNameCol is shorter than nameStr.
    !! 
    !!     nameStr = stringCell(catSrcNameCol, catRow)
    !!     call setStringCell(sumCatSrcNameCol, sumRow, nameStr)
!! !print *, 'sumRow, catRow, sumSrcNumData(sumRow), catSrcNumData(catRow): '&
!! !, sumRow, catRow, sumSrcNumData(sumRow), catSrcNumData(catRow)
!! !print *, 'nameStr: ', trim(nameStr)
    !!   end if
    !! end do

    call release(sumSet)	! Cause clobberFailed if clobber().
    ! deallocate(catObsIdDataInt)
    call release(catSet)

  end subroutine ingestsrcnames


  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

    ArgCatSetTabName     = stringParameter('cattab')
    ArgSrcListSetTabName = stringParameter('srclisttab')
    ArgWithNewSrcListSet = booleanParameter('withnewsrclist')
    if (ArgWithNewSrcListSet) then
      ArgNewSrcListSetName   = stringParameter('newsrclistset')
    end if

    ArgWithObsId = booleanParameter('withobsid')
    if (ArgWithObsId) then
      ArgObsId   = intParameter('obsid')
    end if

  end subroutine readParameters

end module ingestsrcnames_mod
