package ThumbTest;
# ESA (C) 2000-2018 
#
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

# Author: Masaaki Sakano - SSC/LUX
# NOTE: Most of subroutines should go into SSCLib.pm once ssclib is sorted out.

our $VERSION = '1.0.0';

use strict;
use Data::Dumper;
use Exporter 'import';
our @EXPORT = qw(construct_commandstr getarcom1element_for_sasarg get_thumb_out_prms);             # symbols to export in default
# our @EXPORT_OK = qw();  # symbols to export on request
# Not-Exported: 


#---------------------------------------------------------------------
sub construct_commandstr {
#---------------------------------------------------------------------
  ## Input: (Command-name, \hash, [\@arg_order])
  ##   An optional argument @arg_order is a _reference_ to Array.
  ##   Note that if accidentally a normal array is given, it will
  ##   raise an error, like <Can't use string ("-a") as an ARRAY ref>.
  ##   If given, the order of the arguments in the returned string
  ##   follows the order in the array.  In that case, all the possible
  ##   keys for the arguments have to appear in @arg_order .
  ##   On the contrary, if not given, the order is arbitrary.
  ## Return: String (=all the command-line arguments)
  ## Description: Return the command-line string for SAS tasks.
  ## Examples: 
  ##   my(@d,%hs,@arhs);
  ##   @d = qw(--noclobber -V set);
  ##   %hs      = ("-V"=>4, "set"=>"xyz.ds");
  ##   $arhs[1] = {"-V"=>6, "nonexistent"=>"not_cause_error_in_Sub"};
  ##   $stat0=&SSCLib::run(&construct_commandstr("sas",\%hs,    \%d));
  ##   $stat1=&SSCLib::run(&construct_commandstr("sas",\%{$arhs[1]}));
  ## Reference: 

  my $comname = shift;	# e.g., "makethumbs -V 6"
  my $arg = shift;
  my %hscom = %$arg;	# dereference
  my @arg_order = ();
  if (scalar(@_) >= 1) {
    $arg = shift;
    @arg_order = @$arg;
  }
  my(@arcom) = ($comname);
  chomp($arcom[0]);
  my($key, $value, $s);

  if (scalar(@arg_order) == 0) {
    while (($key,$value) = each %hscom){ 
      push(@arcom, &getarcom1element_for_sasarg($key, $value, $comname));
    }

  } else {
    foreach $key (@arg_order) {
      if (exists($hscom{$key})) {
        push(@arcom, &getarcom1element_for_sasarg($key, $hscom{$key}, $comname));
      }
    }

    # Check
    foreach $key (keys(%hscom)) {
      $s=quotemeta($key);
      if ( ! grep(/^$s$/, @arg_order) ){
	die "Undefined key ($key) in \@arg_order is found for ($comname).  Contact the code developer.";
      }
    }
  }

  return(join(" ", @arcom));
}


#---------------------------------------------------------------------
sub getarcom1element_for_sasarg {
#---------------------------------------------------------------------
  ## Input: (Key, Value, Command_name)
  ## Return: String to be used 
  ## Description: Return an element of the command-line argument
  ##    for a SAS task.
  ## Reference: 

  my $key = shift;
  my $value = shift;
  my $comname = shift;
  my $retstr;

  if ($key =~ /^\s*$/) {
    die "Empty key is found for ($comname).  Contact the code developer.";
  } elsif ((! defined($value)) || ($value =~ /^\s*$/)){
    $retstr = $key;			# e.g., --noclobber
  } elsif ($key =~ /^\s*-/) {
    $retstr = $key . " " . $value;	# e.g., -V 6
  } else  {
    $retstr = $key . "=" . $value;	# e.g., set=image.ds
  }
  $retstr =~ s/^\s*//;
  $retstr =~ s/\s*$//;

  $retstr;
}


#---------------------------------------------------------------------
sub get_thumb_out_prms {
#---------------------------------------------------------------------
  ## Input: (string_or_array_of_stdout_from_command)
  ## Return: Array of hashes of Thumbnail output parameters
  ## Description: 
  ## USAGE: @a=&get_thumb_out_prms(@ar=<COMMAND_OUTPUT>); %r=%$a[2];
  ## Reference: 

  my @arin = @_;
  map {chomp; $_.="\n";} @arin;
  if (scalar(@arin) == 1) {
    @arin = split(/\n/, $arin[0]);
  }
  map {chomp;} @arin;
  my %hsdef = (
    'filename'     => undef,
    'radeg'        => undef,
    'decdeg'       => undef,
    'srcindex'     => undef,
    'indexstr'     => undef,
    'fullindexstr' => undef,
    'detid'        => undef,
    'srcid'        => undef,
    'xpixel'       => undef,
    'ypixel'       => undef,
    'catname'      => undef,
    'fullcatname'  => undef,
    'incatflag'    => undef
  );
  my @ardefkey = map {lc} keys(%hsdef);
  my %hslast = %hsdef;
  my %hstmp;
  my(@arlastlines) = ();
  my(@arretall, @arret);
  my(@arsplit, @arsplitkey, @arsplitval);
  my($eachline, $eachc, $eachret, $eachs, $eachcand, $strkeycand, $strvalcand);
  my($i, $nsplit, $str, $s, $slc, $qs, $flagin);

  ####### An example of the input strings #######
  #
  # makethumbs:- Created GIF file is:
  # outfiles/tmpmakethumbs_test_all_15_3.GIF
  #   filename: (outfiles/tmpmakethumbs_test_all_15_3.GIF)
  #   (raDeg, decDeg): (  86.5332288184136331 ,  -69.1672784938972995 )
  #   srcIndex:  3
  #   indexStr:     (069872)
  #   fullIndexStr: (:069872)
  #   detid:     69872
  #   srcid:     53956
  #   (xPixel, yPixel): (   3.4975909E+02 ,    3.7673938E+02 )
  #   catName:     (2XMM J054607.9-691002)
  #   fullCatName: (2XMM J054607.9-691002)
  #   inCatFlag:  T
  #  ====================
  # makethumbs:- makethumbs (src_thumbs-0.2)  [???] ended:    2009-10-14T13:17:48.000

  $flagin = 0;
  foreach $eachline (@arin) {
    if ($eachline =~ /^\s+filename:\s+\(/) {
      $flagin = 1;	# Start storing the parameters
      @arlastlines = ($eachline);	# Initialisation

    } elsif ($flagin && ($eachline =~ /^\s+={5,}\s*$/)) {
      $flagin = 0;	# End storing the parameters

      if ((scalar(@arlastlines) < 2) || (scalar(@arlastlines) > 12)) {	## Sanity check
	warn " =========== Something is wrong in &get_thumb_out_prms().";
	print Dumper \@arlastlines;
	die "FATAL: Can not continue.";
      }

      foreach $eachcand (@arlastlines) {
	next if ($eachcand =~ /^\s*$/);

	if ($eachcand =~ /^\s+([^:]+):\s*(.+)$/) {
	  $strkeycand = $1;
	  $strvalcand = $2; 
	  $strkeycand =~ s/^\s*\(\s*//;	# Remove parentheses
	  $strkeycand =~ s/\s*\)\s*$//;	#  => "raDeg, decDeg" etc
	  $strvalcand =~ s/^\s*\(\s*//;	# Remove parentheses
	  $strvalcand =~ s/\s*\)\s*$//;	#  => "1.23 , 4.56" etc

	  @arsplitkey = split(/\s*,\s*/, $strkeycand);
	  $nsplit = scalar(@arsplitkey);
	  if ($nsplit == 1) {
	    # e.g., catName:     (2XMM J054607.9-691002)
	    $slc = lc($arsplitkey[0]);
	    $qs = quotemeta($slc);
	    if (grep(/^${qs}$/, @ardefkey)) {
	      $hslast{$slc} = $strvalcand;	# filename, detid etc.
	    } else {
	      print Dumper \@ardefkey;
	      die "FATAL: No key ($arsplitkey[0]) in Line ($eachcand)"; 
	    }

          } elsif ($nsplit == 2) { 
	    # e.g., (xPixel, yPixel): (   3.4975909E+02 ,    3.7673938E+02 )
	    @arsplitval = split(/\s*,\s*/, $strvalcand);
	    if (scalar(@arsplitval) != 2) {
	      die "FATAL: Value has only one value where two required in Line ($eachcand)"; 
	    }
	    $hslast{lc($arsplitkey[0])} = $arsplitval[0];	# xPixel etc
	    $hslast{lc($arsplitkey[1])} = $arsplitval[1];	# yPixel etc

	  } else {	# $nsplit = scalar(@arsplitkey); if ($nsplit == 1) {
	    die "FATAL: Line looks strange -- more than 2 components ($eachcand)"; 
	  }
    
	} else {	# if ($eachcand =~ /^\s+([^:]+):\s*(\S+)\s*$/) {
	  ## Namely, there is no colon ":" in the line.
	  print Dumper \@arsplitkey;
	  die "FATAL: Line looks strange (nsplit=$nsplit) ($eachcand)"; 
	}

      }	# foreach $eachcand (@arlastlines) {

      $arretall[$#arretall+1] = { %hslast };

      %hslast = %hsdef;	# Initialisation

    } elsif ($flagin) {
      push(@arlastlines, $eachline);
    }	# if ($eachline =~ /^\s+filename:\s+\(/) {
  }	# foreach $eachline (@arin) {

  ## Remove duplication from @arretall
  @arret = ();
  LOOP_ARRETALL: foreach $eachc (@arretall) {
    %hstmp = %$eachc;
    next if ($hstmp{'filename'} eq '');	# Ignore when output Filename is not defined, which may happen when Noisy.
    foreach $i (0..$#arret) {
      if ($arret[$i]{'filename'} eq $hstmp{'filename'}) {
	$arret[$i] = $eachc;	# Replace with the parameters which appeared later.
	next LOOP_ARRETALL;
      }
    }
    push(@arret, $eachc);	# Add. (n.b., $eachc is a reference to hash.)
  }

  @arret;
}



1;
